#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import random
import shutil
from contextlib import contextmanager
from dataclasses import dataclass
from enum import auto
from typing import List
import numpy as np
import pytest
from pandas import RangeIndex, Series, DataFrame
from frocket.common.dataset import DatasetPartsInfo, DatasetId, DatasetPartId, PartNamingMethod, DatasetInfo, \
    DatasetColumnType, DatasetShortSchema
from frocket.common.serializable import AutoNamedEnum
from frocket.worker.runners.part_loader import shared_part_loader
from tests.utils.base_test_utils import temp_filename, TEMP_DIR, DisablePyTestCollectionMixin
from tests.utils.mock_s3_utils import SKIP_S3_TESTS, new_mock_s3_bucket


class TestColumn(DisablePyTestCollectionMixin, str, AutoNamedEnum):
    int_64_userid = auto()
    int_64_ts = auto()
    int_u32 = auto()
    float_64_ts = auto()
    float_all_none = auto()
    float_32 = auto()
    float_category = auto()
    str_userid = auto()
    str_and_none = auto()
    str_all_none = auto()
    str_object_all_none = auto()
    str_category_userid = auto()
    str_category_few = auto()
    str_category_many = auto()
    bool = auto()
    unsupported_datetimes = auto()
    unsupported_lists = auto()


DEFAULT_GROUP_COUNT = 200
DEFAULT_ROW_COUNT = 1000
DEFAULT_GROUP_COLUMN = TestColumn.int_64_userid.value
DEFAULT_TIMESTAMP_COLUMN = TestColumn.int_64_ts.value
BASE_TIME = 1609459200000  # Start of 2021, UTC
BASE_USER_ID = 100000
TIME_SHIFT = 10000
UNSUPPORTED_COLUMN_DTYPES = {TestColumn.unsupported_datetimes: 'datetime64[ns]',
                             TestColumn.unsupported_lists: 'object'}
STR_AND_NONE_VALUES = ["1", "2", "3"]
STR_CAT_FEW_WEIGHTS = [0.9, 0.07, 0.02, 0.01]
STR_CAT_MANY_WEIGHTS = [0.5, 0.2] + [0.01] * 30


def test_colname_to_coltype(name: str) -> DatasetColumnType:
    prefix_to_type = {
        'int': DatasetColumnType.INT,
        'float': DatasetColumnType.FLOAT,
        'str': DatasetColumnType.STRING,
        'bool': DatasetColumnType.BOOL,
        'unsupported': None
    }
    coltype = prefix_to_type[name.split('_')[0]]
    return coltype


def datafile_schema(part: int = 0) -> DatasetShortSchema:
    # noinspection PyUnresolvedReferences
    result = DatasetShortSchema(
        min_timestamp=float(BASE_TIME),
        max_timestamp=float(BASE_TIME + TIME_SHIFT),
        source_categoricals=[TestColumn.str_category_userid, TestColumn.str_category_many],
        potential_categoricals=[TestColumn.str_and_none, TestColumn.str_category_few],
        columns={col.value: test_colname_to_coltype(col)
                 for col in TestColumn
                 if test_colname_to_coltype(col)})

    # print(f"Test dataset short schema is:\n{result.to_json(indent=2)}")
    return result


def weighted_list(size: int, weights: list) -> list:
    res = []
    for idx, w in enumerate(weights):
        v = str(idx)
        vlen = size * w
        res += [v] * int(vlen)
    assert len(res) == size
    return res


def str_and_none_column_values(part: int = 0, with_none: bool = True) -> List[str]:
    result = [*STR_AND_NONE_VALUES, f"part-{part}"]
    if with_none:
        result.append(None)
    return result


def create_datafile(part: int = 0, size: int = DEFAULT_ROW_COUNT, filename: str = None) -> str:
    # First, prepare data for columns

    # Each part has a separate set of user (a.k.a. group) IDs
    initial_user_id = BASE_USER_ID * part
    min_user_id = initial_user_id
    max_user_id = initial_user_id + DEFAULT_GROUP_COUNT - 1
    # To each tests, ensure that each user ID appears in the file at least once, by including the whole range,
    # then add random IDs in the range
    int64_user_ids = \
        list(range(min_user_id, max_user_id + 1)) + \
        random.choices(range(min_user_id, max_user_id + 1), k=size - DEFAULT_GROUP_COUNT)
    # And also represent as strings in another column
    str_user_ids = [str(uid) for uid in int64_user_ids]

    # Timestamp: each part has a range of values of size TIME_SHIFT
    min_ts = BASE_TIME + (TIME_SHIFT * part)
    max_ts = BASE_TIME + (TIME_SHIFT * (part + 1))
    # Ensure that min & max timestamps appear exactly once, and fill the rest randomly in the range
    int_timestamps = \
        [min_ts, max_ts] + \
        random.choices(range(min_ts + 1, max_ts), k=size-2)
    # Now as floats and as (incorrect!) datetimes (datetimes currently unsupported)
    float_timestamps = [ts + random.random() for ts in int_timestamps]

    # More test columns
    int_u32_values = random.choices(range(100), k=size)
    float_32_values = [np.nan, *[random.random() for _ in range(size - 2)], np.nan]
    str_and_none_values = random.choices(str_and_none_column_values(part), k=size)
    bool_values = random.choices([True, False], k=size)

    # For yet-unsupported columns below
    lists_values = [[1, 2, 3]] * size
    datetimes = [ts * 1000000 for ts in float_timestamps]

    # Now create all series
    idx = RangeIndex(size)
    columns = {
        TestColumn.int_64_userid: Series(data=int64_user_ids),
        TestColumn.int_64_ts: Series(data=int_timestamps),
        TestColumn.int_u32: Series(data=int_u32_values, dtype='uint32'),
        TestColumn.float_64_ts: Series(data=float_timestamps),
        TestColumn.float_all_none: Series(data=None, index=idx, dtype='float64'),
        TestColumn.float_32: Series(data=float_32_values, dtype='float32'),
        TestColumn.float_category: Series(data=float_timestamps, index=idx, dtype='category'),
        TestColumn.str_userid: Series(data=str_user_ids),
        TestColumn.str_and_none: Series(data=str_and_none_values),
        TestColumn.str_all_none: Series(data=None, index=idx, dtype='str'),
        TestColumn.str_object_all_none: Series(data=None, index=idx, dtype='object'),
        TestColumn.str_category_userid: Series(data=str_user_ids, dtype='category'),
        TestColumn.str_category_few: Series(data=weighted_list(size, STR_CAT_FEW_WEIGHTS)),
        TestColumn.str_category_many: Series(data=weighted_list(size, STR_CAT_MANY_WEIGHTS), dtype='category'),
        TestColumn.bool: Series(data=bool_values, dtype='bool'),
        TestColumn.unsupported_datetimes: Series(data=datetimes, dtype='datetime64[us]'),
        TestColumn.unsupported_lists: Series(data=lists_values)
    }
    assert sorted([e for e in columns.keys()]) == sorted([col for col in TestColumn])  # All enum members covered!

    df = DataFrame({col.value: values for col, values in columns.items()})
    if not filename:
        filename = temp_filename('.testpq')
    df.to_parquet(filename)
    # print(f"Written DataFrame to {filename}, columns:\n{df.dtypes}\nSample:\n{df}")
    return filename


create_datafile()


@pytest.fixture(scope="module")
def datafile() -> str:
    return create_datafile()


@dataclass
class TestDatasetInfo(DisablePyTestCollectionMixin):
    basepath: str
    basename_files: List[str]
    fullpath_files: List[str]
    expected_parts: DatasetPartsInfo
    default_dataset_id: DatasetId
    default_dataset_info: DatasetInfo
    bucket: object = None

    def expected_part_ids(self, dsid: DatasetId, parts: List[int] = None) -> List[DatasetPartId]:
        parts = parts or range(len(self.fullpath_files))
        return [DatasetPartId(dataset_id=dsid, path=self.fullpath_files[i], part_idx=i)
                for i in parts]

    # noinspection PyUnresolvedReferences
    def copy_to_s3(self, path_in_bucket: str = '') -> str:
        assert not SKIP_S3_TESTS
        if not self.bucket:
            self.bucket = new_mock_s3_bucket()

        for i, file_fullpath in enumerate(self.fullpath_files):
            file_basename = self.basename_files[i]
            if path_in_bucket:
                file_basename = f"{path_in_bucket}/{file_basename}"
            self.bucket.upload_file(file_fullpath, file_basename)

        s3path = f"s3://{self.bucket.name}/{path_in_bucket}"
        return s3path

    def make_part(self, part_idx: int, s3path: str = None) -> DatasetPartId:
        if not s3path:
            path = self.fullpath_files[part_idx]
        else:
            if not s3path.endswith('/'):
                s3path += '/'
            path = f"{s3path}{self.basename_files[part_idx]}"
        return DatasetPartId(dataset_id=self.default_dataset_id, path=path, part_idx=part_idx)

    # noinspection PyUnresolvedReferences
    def cleanup(self):
        assert self.basepath.startswith(TEMP_DIR)
        shutil.rmtree(self.basepath)
        if self.bucket:
            for key in self.bucket.objects.all():
                key.delete()
            self.bucket.delete()

    @staticmethod
    def build(parts: int, prefix: str = '', suffix: str = '.parquet'):
        basepath = temp_filename(suffix="-dataset")
        os.makedirs(basepath)
        basename_files = sorted([f"{prefix}{temp_filename(suffix=('-p' + str(i) + suffix), with_dir=False)}"
                                 for i in range(parts)])
        fullpath_files = [f"{basepath}/{fname}" for fname in basename_files]
        [create_datafile(part=i, filename=fname) for i, fname in enumerate(fullpath_files)]

        expected_parts = DatasetPartsInfo(naming_method=PartNamingMethod.LIST,
                                          total_parts=parts,
                                          total_size=sum([os.stat(fname).st_size for fname in fullpath_files]),
                                          running_number_pattern=None,
                                          filenames=basename_files)

        dsid = DatasetId.now(name='dsid:' + basepath)
        # noinspection PyTypeChecker
        ds = DatasetInfo(id=dsid,
                         basepath=basepath,
                         total_parts=parts,
                         group_id_column=DEFAULT_GROUP_COLUMN,
                         timestamp_column=DEFAULT_TIMESTAMP_COLUMN)
        return TestDatasetInfo(basepath=basepath, basename_files=basename_files,
                               fullpath_files=fullpath_files, expected_parts=expected_parts,
                               default_dataset_id=dsid, default_dataset_info=ds)


@contextmanager
def new_test_dataset(parts: int, prefix: str = '', suffix: str = '.parquet') -> TestDatasetInfo:
    test_ds = TestDatasetInfo.build(parts=parts, prefix=prefix, suffix=suffix)
    try:
        yield test_ds
    finally:
        test_ds.cleanup()


def are_test_dfs_equal(df1: DataFrame, df2: DataFrame) -> bool:
    # Pandas doesn't know how to compare object-type columns holding more than a single object,
    # so ignore that column
    cols_to_ignore = list(UNSUPPORTED_COLUMN_DTYPES.keys())
    diff_df = df1.drop(columns=cols_to_ignore, errors='ignore').compare(
        df2.drop(columns=cols_to_ignore, errors='ignore'))
    return len(diff_df) == 0


# noinspection PyProtectedMember
@contextmanager
def clean_loader_cache(size_mb: float = None) -> None:
    loader = shared_part_loader()
    try:
        loader._setup(size_mb)
        yield None
    finally:
        loader._setup()  # Cleanup and make ready for next use



"""
Run the dataset registration task for a single part.

The task's job is to validate *a single file only*, return its validity & schame, and optionally return the list of
unique group IDs in the files so that the job object (on the invoker side) can then validate schema & uniques ACROSS
the sampled files.
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import math
from typing import cast, Optional
import numpy as np
from pandas import DataFrame, Series, CategoricalDtype
from pandas.core.dtypes.common import is_numeric_dtype, is_bool_dtype, is_integer_dtype, \
    is_float_dtype, is_string_dtype, is_categorical_dtype
from frocket.common.config import config
from frocket.common.dataset import DatasetSchema, DatasetColumn, DatasetColumnType, DatasetColumnAttributes
from frocket.common.metrics import MetricName
from frocket.common.tasks.base import TaskStatus, BaseTaskRequest, TaskAttemptId, BlobId
from frocket.common.tasks.registration import RegistrationTaskResult, RegistrationTaskRequest
from frocket.common.helpers.utils import ndarray_to_bytes
from frocket.common.validation.consts import CONDITION_COLUMN_PREFIX
from frocket.worker.runners.base_task_runner import BaseTaskRunner, TaskRunnerContext

logger = logging.getLogger(__name__)
# The top values for categoricals as recorded in the schema are approximate, as we're not reading all dataset files
# to get exact counts, nor return a complete list of value->counts per file. Instead, there's a simple factor over the
# amount of configured top values length which each task returns.
TOP_GRACE_FACTOR = 1.5


class RegistrationTaskRunner(BaseTaskRunner):
    def __init__(self, req: BaseTaskRequest, ctx: TaskRunnerContext):

        super().__init__(req, ctx)
        self._req = cast(RegistrationTaskRequest, req)
        self._part_id = self._req.part_id
        self._task_attempt_id = TaskAttemptId(req.invoker_set_task_index, req.attempt_no)  # TODO backlog handle nicer
        self._schema = None
        self._uniques_blob_id = None
        self._warnings = []
        # See configuration guide for these values
        # TODO backlog pass these configuration attrs from the invoker, so they can be controlled there
        self._categorical_ratio = config.float('dataset.categorical.ratio')
        self._categorical_top_count = math.floor(config.int('dataset.categorical.top.count') * TOP_GRACE_FACTOR)

    def _do_run(self):
        self._update_status(TaskStatus.LOADING_DATA)
        with self._ctx.metrics.measure(MetricName.TASK_TOTAL_LOAD_SECONDS):
            df = self._ctx.part_loader.load_dataframe(self._part_id, self._ctx.metrics)

        self._update_status(TaskStatus.RUNNING)
        self._validate_basic_cols(df)
        self._schema = self._build_schema(df)
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Schema: {self._schema.to_json(indent=2)}")

        if self._req.dataset.group_id_column not in self._schema.columns or \
           self._req.dataset.timestamp_column not in self._schema.columns:
            raise Exception(f"Group column or timestamp column not included in result schema: {self._schema.to_json()}")

        if self._req.return_group_ids:
            self._uniques_blob_id = self.write_uniques_as_blob(df)
            logger.info(f"Wrote unique group IDs to blob ID: {self._uniques_blob_id}")

    def _validate_basic_cols(self, df: DataFrame):
        # Validate presence & content of the group ID & timestamp columns, which are mandatory in any dataset
        col_names = df.dtypes.index.tolist()
        group_id_colname = self._req.dataset.group_id_column
        ts_colname = self._req.dataset.timestamp_column

        if group_id_colname not in col_names:
            raise Exception(f"Column {group_id_colname} not found in file")
        if not (is_integer_dtype(df[group_id_colname]) or is_string_dtype(df[group_id_colname])):
            raise Exception(f"Group column {group_id_colname} type must be either integer or string, "
                            f"but found: {df[group_id_colname].dtype}")
        if df[group_id_colname].isnull().any():
            raise Exception(f"Some values in column {group_id_colname} are null")
        logger.info(f"Column {group_id_colname} was found, and without null values")

        if ts_colname not in col_names:
            raise Exception(f"Column {ts_colname} not found in file")
        ts_column = df[ts_colname]
        if not is_numeric_dtype(ts_column.dtype):
            raise Exception(f"Column {ts_colname} is not numeric")
        if np.isnan(ts_column).any():
            raise Exception(f"Column {ts_colname} has NaN values")
        logger.info(f"Column {ts_colname} was found, is numeric and without NaN values")

    def _build_schema(self, df: DataFrame) -> DatasetSchema:
        columns = {}
        unsupported_columns = {}
        all_columns = {name: df[name] for name in df.columns.tolist()}
        for name, series in all_columns.items():
            # noinspection PyBroadException
            try:
                self._validate_column_name(name)
                col_info = self._build_column(series)
                if col_info:
                    columns[name] = col_info
                else:
                    unsupported_columns[name] = str(series.dtype)
            except Exception as e:
                self._warnings.append(f"Skipping column {name} due to error: {e}")
                unsupported_columns[name] = str(series.dtype)

        if len(self._warnings) > 0:
            logger.warning(f"Encountered the following warnings: {self._warnings}")

        return DatasetSchema(columns=columns, unsupported_columns=unsupported_columns,
                             group_id_column=self._req.dataset.group_id_column,
                             timestamp_column=self._req.dataset.timestamp_column)

    # noinspection PyMethodMayBeStatic
    def _validate_column_name(self, name: str) -> None:
        if name.startswith(CONDITION_COLUMN_PREFIX):
            raise Exception(f"Column name starting with {CONDITION_COLUMN_PREFIX} may be overriden by query engine")

    def _build_column(self, series: Series) -> Optional[DatasetColumn]:
        coltype = None
        colattrs = None
        if is_categorical_dtype(series.dtype):
            actual_dtype = series.cat.categories.dtype  # e.g. str for a categorical string column
        else:
            actual_dtype = series.dtype

        if is_bool_dtype(actual_dtype):
            coltype = DatasetColumnType.BOOL
            colattrs = DatasetColumnAttributes()
        elif is_numeric_dtype(actual_dtype):
            if is_integer_dtype(actual_dtype) or is_float_dtype(actual_dtype):
                coltype = DatasetColumnType.INT if is_integer_dtype(actual_dtype) else DatasetColumnType.FLOAT
                stats = series.describe()
                colattrs = DatasetColumnAttributes(numeric_min=stats['min'], numeric_max=stats['max'])
        elif self._is_string_series(series):
            coltype = DatasetColumnType.STRING
            colattrs = self._build_string_attributes(series)

        if coltype:
            return DatasetColumn(name=str(series.name), dtype_name=str(series.dtype),
                                 coltype=coltype, colattrs=colattrs)
        else:
            return None

    @staticmethod
    def _is_string_series(series: Series):
        def string_or_none(v):
            return v is None or type(v) is str

        # If the validated file was created with Pandas, it might already have categorical columns -
        # in which case the .apply() call below also returns a categorical type that will need conversion to bool
        if is_string_dtype(series.dtype) or \
                (is_categorical_dtype(series.dtype) and is_string_dtype(series.cat.categories.dtype)):
            # If the series has a mix of strings and other types, Pandas may think it's a string column -
            # so ensure no non-strings are present.
            # See: https://youtrack.jetbrains.com/issue/PY-43841 for type warning issue below
            # noinspection PyTypeChecker
            are_all_strings = series.apply(string_or_none).astype('bool').all()
            if are_all_strings:
                return True
        return False

    def _build_string_attributes(self, series: Series) -> DatasetColumnAttributes:
        dtype = series.dtype
        is_categorical = is_categorical_dtype(dtype)
        unique_values_ratio = None

        if is_categorical:
            categories = cast(CategoricalDtype, dtype).categories
            # If field was loaded as categorical, respect that. We don't validate the ratio, but do store it.
            unique_values_ratio = len(categories) / len(series)
        else:
            categories = cast(CategoricalDtype, series.astype('category').dtype).categories
            if len(categories) > 0:
                # Only treat string columns as categorical if the ratio of unique values to total values is below X
                unique_values_ratio = len(categories) / len(series)
                if unique_values_ratio <= self._categorical_ratio:
                    is_categorical = True

        if is_categorical:
            if self._categorical_top_count == 0:  # Don't return top values
                top_values = None
            else:
                top_values = series.value_counts(normalize=True)[0:self._categorical_top_count].to_dict()

            return DatasetColumnAttributes(categorical=True,
                                           cat_unique_ratio=unique_values_ratio,
                                           cat_top_values=top_values)
        else:
            return DatasetColumnAttributes()

    def write_uniques_as_blob(self, df) -> BlobId:
        uniques = df[self._req.dataset.group_id_column].unique()
        array_bytes = ndarray_to_bytes(uniques)
        logger.info(f"Uniques array len: {len(uniques)}, serialized bytes: {len(array_bytes)}")

        # Tag is optional, just for debugging/understanding keys' meaning in the blobstore
        blob_tag = f"uniques-{self._part_id.dataset_id.name}-{self._part_id.part_idx}"
        return self._ctx.blobstore.write_blob(array_bytes, tag=blob_tag)

    def _build_result(self, base_attributes: dict):
        return RegistrationTaskResult(**base_attributes,
                                      dataset_schema=self._schema,
                                      part_id=self._part_id,
                                      group_ids_blob_id=self._uniques_blob_id)



"""
Base types for all concrete task and job classes
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import time
from enum import auto
from typing import Optional, Dict, List, NamedTuple
from dataclasses import dataclass, field
from frocket.common.metrics import MetricData
from frocket.common.serializable import SerializableDataClass, AutoNamedEnum

logger = logging.getLogger(__name__)

ErrorMessage = str
BlobId = str


@dataclass(frozen=True)
class BaseTaskRequest(SerializableDataClass):
    """
    All task requests (enqueud by the invoker, processed by a worker) inherit from this class.

    There are two main modes of request:
    1. Either the invoker explicitly specifies a task_index and all the details of the task to run
       (as attributes of the concrete task request class), OR:
    2. (For jobs which support it, where it makes sense to do)
       The invoker publishes a set of tasks to do in the datastore, and then launches the appropriate amount of task
       requests - but it's then the worker role to choose the task to perform. This gives workers a chance
       to choose work over files they already have cached, or fallback to any remaining task.
       See: query_task_runner.py
    """
    request_id: str  # Unique ID for storing all request/response related payloads in the datastore
    invoke_time: float
    attempt_no: int
    invoker_set_task_index: Optional[int]


class TaskStatus(AutoNamedEnum):
    def __init__(self, *args):
        self.ended = args[0].startswith('ENDED_')

    QUEUED = auto()
    LOADING_DATA = auto()
    RUNNING_QUERY = auto()
    RUNNING = auto()  # Generic status for non-query tasks
    ENDED_SUCCESS = auto()
    ENDED_FAILED = auto()


@dataclass(frozen=True)
class BaseTaskResult(SerializableDataClass):
    task_index: int
    status: TaskStatus
    error_message: Optional[str]  # Only (possibly) available if status is ENDED_FAILED
    metrics: List[MetricData]


@dataclass(frozen=True)
class TaskAttemptId(SerializableDataClass):
    """Represents a single attempt at running a specific task_index."""
    task_index: int
    attempt_no: int = 0


@dataclass(frozen=True)
class TaskStatusUpdate(SerializableDataClass):
    """Couple a status change with when it was set."""
    status: TaskStatus
    time: float

    @classmethod
    def now(cls, status):
        return TaskStatusUpdate(status, time.time())


class TaskAttemptsInfo:
    """
    A clumsily-named class used by the invoker to determine the latest status a given task_index.

    The invoker reads all attempt statuses from the datastore, and holds an instance of this object
    for each task_index, calling add() for all attempts for it (hopefully, just one).
    """
    class AttemptAndUpdate(NamedTuple):
        attempt_id: TaskAttemptId
        update: TaskStatusUpdate

    def __init__(self, task_index: int):
        self._task_index = task_index
        self._attempts: Dict[TaskAttemptId, TaskAttemptsInfo.AttemptAndUpdate] = {}
        self._latest: Optional[TaskAttemptsInfo.AttemptAndUpdate] = None

    def add(self, attempt_id: TaskAttemptId, update: TaskStatusUpdate) -> None:
        assert (attempt_id.task_index == self._task_index)
        self._attempts[attempt_id] = TaskAttemptsInfo.AttemptAndUpdate(attempt_id, update)
        self._latest = None  # Invalidate

    def _update_latest_attempt(self):
        """
        Determine latest attempt:

        1. if there is an attempt with ENDED_SUCCESS status - take the earliest attempt with that status.
           Note it IS possible to have more than one, in case an attempt was judged "lost" and another attempt launched.
           If then the first attempt manages to complete - we may have >1 attempts with ENDED_SUCCESS status by the
           point where all tasks have ended. This is fine AS LONG AS ONLY THE RESULT OF A SINGLE ONE IS USED to compose
           the aggregated result, or you will have double+ counts for such tasks...
        2. If there is no attempt with ENDED_SUCCESS status (for now), take the status of the most recent attempt.
        """
        if not self._latest and len(self._attempts) > 0:
            attempt_tuples = self._attempts.values()
            ordered_attempts = sorted(attempt_tuples, key=lambda t: t.attempt_id.attempt_no)

            ended_attempts = [t for t in ordered_attempts if t.update.status == TaskStatus.ENDED_SUCCESS]
            if ended_attempts:
                self._latest = ended_attempts[0]
            else:
                self._latest = ordered_attempts[-1]

    @property
    def task_index(self) -> int:
        return self._task_index

    @property
    def latest_attempt(self) -> Optional[TaskAttemptId]:
        self._update_latest_attempt()
        return self._latest.attempt_id if self._latest else None

    @property
    def latest_update(self) -> Optional[TaskStatusUpdate]:
        self._update_latest_attempt()
        return self._latest.update if self._latest else None

    @property
    def attempt_count(self) -> int:
        return len(self._attempts)

    @property
    def time_since_update(self) -> Optional[float]:
        self._update_latest_attempt()
        return time.time() - self._latest.update.time if self._latest else None

    @property
    def attempts(self) -> Dict[TaskAttemptId, TaskStatusUpdate]:
        return {attempt_id: t.update for attempt_id, t in self._attempts.items()}


@dataclass(frozen=True)
class BaseApiResult(SerializableDataClass):
    """ The minimal result from invoker_api operations which return a status, but not necessarily
    run a job (invokes tasks via workers)."""
    success: bool
    error_message: Optional[str]


@dataclass(frozen=True)
class JobDatasetStats(SerializableDataClass):
    total_size: int
    parts: int


@dataclass(frozen=True)
class JobInvokerStats(SerializableDataClass):
    enqueue_time: float
    poll_time: float
    total_tasks: int
    failed_tasks: int
    task_success_over_time: Dict[float, int]  # Non-cumulative
    # TODO backlog lost_task_retries: int - this should be reported by the invoker as task status doesn't signal it


TimingStats = Dict[str, float]


@dataclass(frozen=True)
class JobWorkerStats(SerializableDataClass):
    cold_tasks: int
    warm_tasks: int
    invoke_latency: TimingStats
    load_time: TimingStats
    total_time: TimingStats
    scanned_rows: int
    scanned_groups: int
    cache: Dict[str, int]
    # TODO backlog loaded_column_types: Dict[str, int]  - type: count of all column types actually loaded,
    #  e.g. int, float, string (and string:categorical!) for optimizing query times.


@dataclass(frozen=True)
class JobStats(SerializableDataClass):
    total_time: Optional[float] = field(default=None)
    invoker: Optional[JobInvokerStats] = field(default=None)
    worker: Optional[JobWorkerStats] = field(default=None)
    dataset: Optional[JobDatasetStats] = field(default=None)
    cost: Optional[float] = field(default=None)


@dataclass(frozen=True)
class BaseJobResult(BaseApiResult):
    """Concrete jobs' result classes inherit from this class."""
    request_id: str
    stats: JobStats


@dataclass(frozen=True)
class JobStatus:
    """Final status of a job run & all its task attempts."""
    success: bool
    error_message: Optional[str]
    attempts_status: List[TaskAttemptsInfo]

    def with_error(self, message: str) -> object:
        """Copy an existing object, but with an error - create a new "failed" version of this immutable object."""
        if self.error_message:
            message = f"{self.error_message}; {message}"
        return JobStatus(success=False,
                         error_message=message,
                         attempts_status=self.attempts_status)



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import abstractmethod
from typing import Optional
from frocket.common.config import config
from frocket.common.tasks.base import BlobId

BLOB_DEFAULT_TTL = config.int('blobstore.default.ttl')
BLOB_MAX_TTL = config.int('blobstore.max.ttl')


class Blobstore:
    """Simple interface for storing and fetching arbitrary binary data, for ephemeral transport over the network.
    The data is assumed to always have a default TTL - it's not a permanent or big data store."""
    @abstractmethod
    def write_blob(self, data: bytes, ttl: int = None, tag: str = None) -> BlobId:
        pass

    @abstractmethod
    def read_blob(self, blobid: BlobId) -> Optional[bytes]:
        pass

    @abstractmethod
    def delete_blob(self, blobid: BlobId) -> bool:
        pass



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import time
from typing import List, cast
from frocket.common.config import config
from frocket.common.dataset import DatasetInfo, DatasetPartId, DatasetPartsInfo, DatasetShortSchema
from frocket.common.metrics import JobTypeLabel, DATASET_LABEL
from frocket.common.tasks.query import PartSelectionMode, QueryTaskRequest, QueryTaskResult, QueryJobResult, QueryResult
from frocket.invoker.jobs.job import Job


class QueryJob(Job):
    def __init__(self, dataset: DatasetInfo, parts: DatasetPartsInfo,
                 short_schema: DatasetShortSchema, query: dict, used_columns: List[str],
                 worker_can_select_part: bool = None):
        self._dataset = dataset
        self._parts = parts
        self._query = query
        self._used_columns = used_columns
        self._paths = parts.fullpaths(parent=dataset)
        self._worker_can_select_part = worker_can_select_part \
            if worker_can_select_part is not None else config.bool('worker.self.select.enabled')
        if config.bool('dataset.categorical.potential.use'):
            self._load_as_categoricals = short_schema.potential_categoricals
        else:
            self._load_as_categoricals = None
        self._labels = {
            JobTypeLabel.QUERY.label_name: JobTypeLabel.QUERY.label_value,
            DATASET_LABEL: self._dataset.id.name
        }

    def parts_info(self):
        return self._parts

    def total_tasks(self):
        return len(self._paths)

    def build_tasks(self):
        if self._worker_can_select_part:
            mode = PartSelectionMode.SELECTED_BY_WORKER
        else:
            mode = PartSelectionMode.SET_BY_INVOKER

        requests = [self._build_task(mode, i) for i in range(self.total_tasks())]
        return requests

    def dataset_parts_to_publish(self):
        if self._worker_can_select_part:
            parts_to_publish = {DatasetPartId(self._dataset.id, path, part_index)
                                for part_index, path in enumerate(self._paths)}
            return parts_to_publish
        else:
            return None

    def build_retry_task(self, attempt_no, task_index):
        return self._build_task(PartSelectionMode.SET_BY_INVOKER,
                                part_index=task_index,
                                attempt_no=attempt_no)

    def _build_task(self, mode: PartSelectionMode, part_index: int, attempt_no: int = 0) -> QueryTaskRequest:
        if mode == PartSelectionMode.SET_BY_INVOKER:
            invoker_set_part = DatasetPartId(dataset_id=self._dataset.id,
                                             path=self._paths[part_index],
                                             part_idx=part_index)
            task_index = part_index
        elif mode == PartSelectionMode.SELECTED_BY_WORKER:
            assert attempt_no == 0
            invoker_set_part = None
            task_index = None
        else:
            raise Exception("Unknown mode {mode}")

        request = QueryTaskRequest(
            request_id=self._request_id,
            invoke_time=time.time(),
            dataset=self._dataset,
            load_as_categoricals=self._load_as_categoricals,
            query=self._query,
            invoker_set_task_index=task_index,
            attempt_no=attempt_no,
            mode=mode,
            invoker_set_part=invoker_set_part,
            used_columns=self._used_columns)
        return request

    def build_result(self, base_attributes, final_status, latest_task_results):
        aggregated_query_result = None
        # Only if query was successful, aggregate query results (for each task - from a single successful attempt)
        if final_status.success:
            latest_task_results = cast(List[QueryTaskResult], latest_task_results)
            query_results = [task_result.query_result for task_result in latest_task_results]
            aggregated_query_result = cast(QueryResult,
                                           QueryResult.reduce(query_results))

        result = QueryJobResult(
            **base_attributes,
            query=aggregated_query_result.query if aggregated_query_result else None,
            funnel=aggregated_query_result.funnel if aggregated_query_result else None
        )
        return result



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import time
from typing import cast, List, Union, Tuple
import pytest
from frocket.common.dataset import DatasetPartId
from frocket.common.helpers.utils import timestamped_uuid
from frocket.common.metrics import MetricName, LoadFromLabel, PartSelectMethodLabel, MetricsBag, ComponentLabel
from frocket.common.tasks.base import TaskStatus
from frocket.common.tasks.query import QueryTaskRequest, PartSelectionMode, QueryTaskResult, AggregationResult
from frocket.common.validation.query_validator import QueryValidator
from frocket.datastore.registered_datastores import get_datastore
from frocket.worker.runners.base_task_runner import DEFAULT_PREFLIGHT_DURATION_MS
from frocket.worker.runners.part_loader import PartLoader
from tests.utils.base_test_utils import assert_metric_value, assert_label_value_exists, find_first_label_value, \
    get_metric_value
from tests.utils.dataset_utils import new_test_dataset, str_and_none_column_values, TestDatasetInfo, TestColumn, \
    DEFAULT_ROW_COUNT, DEFAULT_GROUP_COUNT, DEFAULT_GROUP_COLUMN, DEFAULT_TIMESTAMP_COLUMN, BASE_TIME, TIME_SHIFT, \
    datafile_schema
from tests.utils.mock_s3_utils import SKIP_S3_TESTS
from tests.utils.task_and_job_utils import simple_run_task

# Note: not trying to test the query engine itself here, but that the query task "wraps it" correctly.
# Specifically, that part self-selection works (in PartSelectionMode.SELECTED_BY_WORKER)

TEST_QUERY_WITH_AGGRS = {
    'query': {
        'conditions': [
            {
                "filter": [
                    DEFAULT_TIMESTAMP_COLUMN,
                    ">=",
                    BASE_TIME
                ]
            }
        ],
        'aggregations': [
            {'column': TestColumn.str_and_none, 'name': 'cond_aggr1', 'type': 'countPerValue'},
            {'column': TestColumn.str_and_none, 'name': 'cond_aggr2', 'type': 'count'}
        ]
    },
    'funnel': {
        "sequence": [
            {
                "filter": [
                    DEFAULT_TIMESTAMP_COLUMN,
                    ">=",
                    BASE_TIME
                ]
            }
        ],
        "stepAggregations": [
            {'column': TestColumn.str_and_none, 'name': 'funnel_step_aggr1', 'type': 'countPerValue'},
            {'column': TestColumn.str_and_none, 'name': 'funnel_step_aggr2', 'type': 'count'}

        ],
        "endAggregations": [
            {'column': TestColumn.str_and_none, 'name': 'funnel_end_aggr1', 'type': 'countPerValue'},
            {'column': TestColumn.str_and_none, 'name': 'funnel_end_aggr2', 'type': 'count'}

        ]
    }
}


def build_task_requests(test_ds: TestDatasetInfo,
                        query: dict,
                        part_selection_mode: PartSelectionMode,
                        parts: List[int],
                        copy_to_s3: bool = False,
                        existing_s3path: str = None,
                        attempt_no: int = 0) -> Tuple[List[QueryTaskRequest], List[DatasetPartId]]:
    if part_selection_mode == PartSelectionMode.SELECTED_BY_WORKER:
        assert attempt_no == 0  # Prevent errors in the test code (forcing an unsupported feature)
    validation_result = QueryValidator(query,
                                       dataset=test_ds.default_dataset_info,
                                       short_schema=datafile_schema()).expand_and_validate()
    assert validation_result.success
    if existing_s3path:
        s3path = existing_s3path
    elif copy_to_s3:
        s3path = test_ds.copy_to_s3()
    else:
        s3path = None

    requests = []
    part_ids = []
    request_id = timestamped_uuid()
    for i, part_idx in enumerate(parts):
        part_id = test_ds.make_part(part_idx, s3path)
        if part_selection_mode == PartSelectionMode.SET_BY_INVOKER:
            invoker_set_part = part_id
            task_index = i
        else:
            invoker_set_part = None
            task_index = None

        req = QueryTaskRequest(
            dataset=test_ds.default_dataset_info,
            query=validation_result.expanded_query,
            mode=part_selection_mode,
            invoker_set_part=invoker_set_part,
            used_columns=[DEFAULT_GROUP_COLUMN, DEFAULT_TIMESTAMP_COLUMN, TestColumn.str_and_none.value],
            load_as_categoricals=None,
            attempt_no=attempt_no, invoke_time=time.time(), request_id=request_id, invoker_set_task_index=task_index)
        requests.append(req)
        part_ids.append(part_id)
    return requests, part_ids


def run_query_task(test_ds: TestDatasetInfo,
                   query: dict,
                   part_selection_mode: PartSelectionMode = PartSelectionMode.SET_BY_INVOKER,
                   parts: Union[int, List[int]] = 0,
                   private_part_loaders: List[PartLoader] = None,
                   copy_to_s3: bool = False,
                   existing_s3path: str = None,
                   preflight_duration_ms: int = None,
                   attempt_no: int = 0) -> Union[List[QueryTaskResult], QueryTaskResult]:
    single_mode = type(parts) is not list
    if single_mode:
        parts = [parts]
    task_requests, part_ids = build_task_requests(test_ds, query, part_selection_mode, parts,
                                                  copy_to_s3, existing_s3path)
    reqid = task_requests[0].request_id
    if part_selection_mode == PartSelectionMode.SELECTED_BY_WORKER:
        get_datastore().publish_for_worker_selection(reqid, attempt_no, set(part_ids))

    results = []
    for i, req in enumerate(task_requests):
        loader = private_part_loaders[i] if private_part_loaders else None
        result = cast(QueryTaskResult, simple_run_task(req, QueryTaskResult, loader, preflight_duration_ms))
        assert result.status == TaskStatus.ENDED_SUCCESS
        results.append(result)

    if part_selection_mode == PartSelectionMode.SELECTED_BY_WORKER:
        assert get_datastore().self_select_part(reqid, attempt_no) is None  # No more parts to select from
    get_datastore().cleanup_request_data(reqid)
    return results[0] if single_mode else results


def test_empty_query():
    with new_test_dataset(1) as test_ds:
        task_result = run_query_task(test_ds, query={}, parts=0,
                                     part_selection_mode=PartSelectionMode.SET_BY_INVOKER)
        assert task_result.query_result.query.matching_groups == DEFAULT_GROUP_COUNT
        assert task_result.query_result.query.matching_group_rows == DEFAULT_ROW_COUNT
        assert_metric_value(task_result.metrics, MetricName.SCANNED_GROUPS, DEFAULT_GROUP_COUNT)
        assert_metric_value(task_result.metrics, MetricName.SCANNED_ROWS, DEFAULT_ROW_COUNT)
        assert_label_value_exists(task_result.metrics, LoadFromLabel.SOURCE)
        assert_label_value_exists(task_result.metrics, PartSelectMethodLabel.SET_BY_INVOKER)
        assert task_result.query_result.funnel is None
        assert task_result.query_result.query.aggregations is None


def test_local_invoker_set_part():
    num_parts = 3
    with new_test_dataset(num_parts) as test_ds:
        task_result = run_query_task(test_ds, TEST_QUERY_WITH_AGGRS,
                                     parts=num_parts - 1,
                                     part_selection_mode=PartSelectionMode.SET_BY_INVOKER)
        assert task_result.query_result.query.matching_groups == DEFAULT_GROUP_COUNT
        assert task_result.query_result.query.matching_group_rows == DEFAULT_ROW_COUNT
        assert len(task_result.query_result.funnel.sequence) == 1
        assert task_result.query_result.funnel.sequence[0].matching_groups == DEFAULT_GROUP_COUNT
        assert task_result.query_result.funnel.sequence[0].matching_group_rows == DEFAULT_ROW_COUNT
        assert_label_value_exists(task_result.metrics, LoadFromLabel.SOURCE)
        assert_label_value_exists(task_result.metrics, PartSelectMethodLabel.SET_BY_INVOKER)

        # Totally hard-coded :-|
        def validate_aggr(aggr_results: List[AggregationResult], expected_name: str):
            assert len(aggr_results) == 2
            aggr1 = aggr_results[0]
            assert aggr1.name == expected_name + '1'
            # By checking the keys in 'str_and_none', test that the right keys for the given part idx exist
            assert aggr1.column == TestColumn.str_and_none and aggr1.type == "countPerValue"
            expected_aggr_keys = sorted(str_and_none_column_values(part=num_parts - 1, with_none=False))
            assert sorted(aggr1.value.keys()) == expected_aggr_keys

            aggr2 = aggr_results[1]
            assert aggr2.name == expected_name + '2'
            assert aggr2.column == TestColumn.str_and_none and aggr2.type == "count"
            assert aggr2.value == sum(aggr1.value.values())

        validate_aggr(task_result.query_result.query.aggregations, expected_name='cond_aggr')
        validate_aggr(task_result.query_result.funnel.sequence[0].aggregations, expected_name='funnel_step_aggr')
        validate_aggr(task_result.query_result.funnel.end_aggregations, expected_name='funnel_end_aggr')

        # Re-run: check that results are ok and that part was again loaded from source (as it is local)
        task_result = run_query_task(test_ds, TEST_QUERY_WITH_AGGRS,
                                     parts=num_parts - 1,
                                     part_selection_mode=PartSelectionMode.SET_BY_INVOKER)
        assert task_result.query_result.query.matching_groups == DEFAULT_GROUP_COUNT
        assert task_result.query_result.query.matching_group_rows == DEFAULT_ROW_COUNT
        assert_label_value_exists(task_result.metrics, LoadFromLabel.SOURCE)


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_s3_invoker_set_part():
    with new_test_dataset(2) as test_ds:
        def validate_counts_and_source(part: int, source: LoadFromLabel):
            task_result = run_query_task(test_ds, query={}, parts=part,
                                         part_selection_mode=PartSelectionMode.SET_BY_INVOKER,
                                         copy_to_s3=True)
            assert task_result.query_result.query.matching_groups == DEFAULT_GROUP_COUNT
            assert task_result.query_result.query.matching_group_rows == DEFAULT_ROW_COUNT
            assert_label_value_exists(task_result.metrics, source)
            assert_label_value_exists(task_result.metrics, PartSelectMethodLabel.SET_BY_INVOKER)

        validate_counts_and_source(0, LoadFromLabel.SOURCE)
        validate_counts_and_source(0, LoadFromLabel.DISK_CACHE)
        validate_counts_and_source(1, LoadFromLabel.SOURCE)
        validate_counts_and_source(1, LoadFromLabel.DISK_CACHE)


def test_worker_set_part_local():
    num_parts = 3
    with new_test_dataset(num_parts) as test_ds:
        def validate():
            task_results = run_query_task(test_ds, query={}, parts=list(range(num_parts)),
                                          part_selection_mode=PartSelectionMode.SELECTED_BY_WORKER,
                                          copy_to_s3=False)
            for res in task_results:
                assert res.query_result.query.matching_groups == DEFAULT_GROUP_COUNT
                assert res.query_result.query.matching_group_rows == DEFAULT_ROW_COUNT
                assert_label_value_exists(res.metrics, LoadFromLabel.SOURCE)
                assert_label_value_exists(res.metrics, PartSelectMethodLabel.RANDOM_NO_CANDIDATES)

        validate()
        # And again - with no caching
        validate()


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_worker_set_part_s3():
    with new_test_dataset(1) as test_ds:
        def validate(s3path: str, source: LoadFromLabel, select_method: PartSelectMethodLabel):
            res = run_query_task(test_ds, query={}, parts=0,
                                 part_selection_mode=PartSelectionMode.SELECTED_BY_WORKER,
                                 existing_s3path=s3path)
            print(res.to_json(indent=2))
            assert res.query_result.query.matching_groups == DEFAULT_GROUP_COUNT
            assert res.query_result.query.matching_group_rows == DEFAULT_ROW_COUNT
            assert_label_value_exists(res.metrics, source)
            assert_label_value_exists(res.metrics, select_method)

        s3path = test_ds.copy_to_s3()
        validate(s3path, LoadFromLabel.SOURCE, PartSelectMethodLabel.RANDOM_NO_CANDIDATES)
        validate(s3path, LoadFromLabel.DISK_CACHE, PartSelectMethodLabel.SPECIFIC_CANDIDATE)


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_self_select_multi_loaders():
    """In worker self-select part mode, validate the expected actual select method:
    Given:
    * 4-part datasets
    * 4 "independent" task runners, each with a separate part loader (meaning, a separate disk cache)
    * The first 3 runners have parts [0, 1] in their cache already.
    Expected: giving each of the runners a chance to run, sequentially, we expect the first two to succesfully select
    eithr part 0 or 1. The third one should report "I have cached candidates, but they're taken already so selected
    a random one (2 or 3), and the last one should say: "I got no candidates, so selected at random".

    In a second run over the same parts, it's expected that the first 3 runner would all successfully select a candidate
    (the first two would take either 0 or 1 again, and the third would take either 2 or 3 - whatever it loaded before),
    and that the 4th and last one would *not* report "I have no candidates" (though it may or may not successfully
    select a candidate rather than pick a random part - as we're dealing with *sets* both at Redis and Python layers)

    We also test that the expected preflight time matches the actual select method: if a task has candidates, it will
    not sleep. If it doesn't, it will sleep till end of preflight. We test this with default and zero preflight values.
    """
    with new_test_dataset(4) as test_ds:
        s3path = test_ds.copy_to_s3()

        def test_for_preflight_duration(preflight_duration_ms: int = None):
            loaders = [PartLoader() for _ in range(4)]
            for loader in loaders[:-1]:
                for part_idx in [0, 1]:
                    metrics = MetricsBag(component=ComponentLabel.WORKER)
                    loader.load_dataframe(test_ds.make_part(part_idx, s3path), metrics)
                    assert_label_value_exists(metrics.finalize(success=True), LoadFromLabel.SOURCE)

            def run():
                task_results = run_query_task(
                    test_ds, query={},
                    part_selection_mode=PartSelectionMode.SELECTED_BY_WORKER,
                    parts=list(range(4)),
                    private_part_loaders=loaders,
                    existing_s3path=s3path,
                    preflight_duration_ms=preflight_duration_ms)

                methods = [find_first_label_value(task_result.metrics, PartSelectMethodLabel)
                           for task_result in task_results]
                preflight_sleeps = [get_metric_value(task_result.metrics, MetricName.TASK_PREFLIGHT_SLEEP_SECONDS)
                                    for task_result in task_results]
                return methods, preflight_sleeps

            actual_methods, sleeps = run()
            expected_methods = [e.label_value for e in [PartSelectMethodLabel.SPECIFIC_CANDIDATE,
                                                        PartSelectMethodLabel.SPECIFIC_CANDIDATE,
                                                        PartSelectMethodLabel.RANDOM_CANDIDATES_TAKEN,
                                                        PartSelectMethodLabel.RANDOM_NO_CANDIDATES]]
            assert sorted(actual_methods) == sorted(expected_methods)
            # Since first 3 had a candidate in mind, they don't spend time in preflight - unlike the last one
            # (unless preflight is disabled - i.e. set to zero)s
            assert sleeps[:-1] == [0, 0, 0]
            if preflight_duration_ms == 0:
                assert sleeps[-1] == 0
            else:
                assert 0 < sleeps[-1] < DEFAULT_PREFLIGHT_DURATION_MS

            # Run again with same loaders. This time all need to have some candidates, even if none
            actual_methods, sleeps = run()
            expected_methods = [PartSelectMethodLabel.SPECIFIC_CANDIDATE.label_value] * 3

            assert sorted(actual_methods[:-1]) == expected_methods
            assert actual_methods[-1] != PartSelectMethodLabel.RANDOM_NO_CANDIDATES.label_value
            assert sleeps == [0] * 4

        test_for_preflight_duration()  # Default
        test_for_preflight_duration(preflight_duration_ms=0)


def test_time_filters():
    with new_test_dataset(1) as test_ds:
        # All rows start at the 'from' value
        q = {"timeframe": {"from": BASE_TIME}}
        res = run_query_task(test_ds, q)
        assert res.query_result.query.matching_group_rows == DEFAULT_ROW_COUNT

        # Only one row should match (have this value)
        q = {"timeframe": {"from": BASE_TIME + TIME_SHIFT}}
        res = run_query_task(test_ds, q)
        assert res.query_result.query.matching_group_rows == 1

        q = {"timeframe": {"from": BASE_TIME, "to": BASE_TIME + 1}}
        res = run_query_task(test_ds, q)
        assert res.query_result.query.matching_group_rows == 1

        q = {"timeframe": {"from": BASE_TIME, "to": BASE_TIME + TIME_SHIFT}}
        res = run_query_task(test_ds, q)
        assert res.query_result.query.matching_group_rows == DEFAULT_ROW_COUNT - 1

        q = {"timeframe": {"from": BASE_TIME, "to": BASE_TIME + TIME_SHIFT + 1}}
        res = run_query_task(test_ds, q)
        assert res.query_result.query.matching_group_rows == DEFAULT_ROW_COUNT

        q = {"timeframe": {"to": BASE_TIME}}
        res = run_query_task(test_ds, q)
        assert res.query_result.query.matching_group_rows == 0

        q = {"timeframe": {"to": BASE_TIME + 1}}
        res = run_query_task(test_ds, q)
        assert res.query_result.query.matching_group_rows == 1



"""
A simple config class based on a dict filled with default values, overridable by env. variables.

There is no support for a config file here, as my personal experience shows managing such files for different
environments tends to be clunky, while env variables are supported by pretty much everything (but YMMV).
Specifically for Funnel Rocket, the env is easily set for AWS Lambdas decoupled from the packaged code.

To override the default value of 'redis.port', set the env variable FROCKET_REDIS_PORT.
See the reference documentation for details of all settings.

TODO backlog switch to a "real" config package OR polish the API and None/empty values handling based on actual usage.
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import logging
import sys
from typing import Dict
import boto3
import botocore

ENV_CONFIG_PREFIX = "FROCKET"
# All default values are specified as strings, just as they may be specified in env variables.
DEFAULTS = {
    "datastore": "redis",
    "blobstore": "redis",
    "redis.host": "localhost",
    "redis.port": "6379",
    "redis.db": "0",
    "datastore.redis.prefix": "frocket:",
    "blobstore.default.ttl": "600",
    "blobstore.max.ttl": "86400",
    "dataset.max.parts": "1000",
    "dataset.categorical.ratio": "0.1",
    "dataset.categorical.top.count": "20",
    "dataset.categorical.top.minpct": "0.01",
    "dataset.categorical.potential.use": "true",
    "part.selection.preflight.ms": "200",
    "invoker": "work_queue",
    "invoker.run.timeout": "60",
    "invoker.async.poll.interval.ms": "25",
    "invoker.async.log.interval.ms": "1000",
    "invoker.lambda.name": "frocket",
    "invoker.lambda.threads": "20",
    "invoker.lambda.debug.payload": "false",
    "invoker.lambda.legacy.async": "true",
    "invoker.retry.max.attempts": "3",
    "invoker.retry.failed.interval": "3",
    "invoker.retry.lost.interval": "20",
    "worker.self.select.enabled": "true",
    "worker.reject.age": "60",
    "metrics.export.prometheus": "true",
    "metrics.export.lastrun": "",
    "metrics.buckets.default": "0.1, 0.5, 1, 5, 25, 100, 1000",
    "metrics.buckets.seconds": "0.05, 0.1, 0.5, 1, 2, 5, 10, 15",
    "metrics.buckets.dollars": "0.01, 0.05, 0.1, 0.5, 1, 2",
    "metrics.buckets.bytes": "1048576, 16777216, 67108864, 134217728, 268435456",
    "metrics.buckets.groups": "100, 1000, 10_000, 100_000, 1_000_000, 10_000_000, 100_000_000, 500_000_000",
    "metrics.buckets.rows":   "100, 1000, 10_000, 100_000, 1_000_000, 10_000_000, 100_000_000, 500_000_000",
    "worker.disk.cache.size.mb": "256",
    "stats.timing.percentiles": "0.25, 0.5, 0.75, 0.95, 0.99",
    "log.file": "",
    "log.level": "info",
    "log.format": "%(asctime)s %(name)s %(levelname)s %(message)s",
    "apiserver.admin.actions": "true",
    "apiserver.error.details": "true",
    "apiserver.prettyprint": "true",
    "validation.sample.max": "10",
    "validation.sample.ratio": "0.1",
    "unregister.last.used.interval": "30",
    "aggregations.top.default.count": "10",
    "aggregations.top.grace.factor": "2.0",
    "aws.endpoint.url": "",
    "aws.access.key.id": "",
    "aws_secret_access_key": "",
    "aws.no.signature": "false",
    "lambda.aws.max.pool.connections": "50",
    "lambda.aws.connect.timeout": "3",
    "lambda.aws.read.timeout": "3",
    "lambda.aws.retries.max.attempts": "3"
}


class ConfigDict(Dict[str, str]):
    def __init__(self, seq=None, **kwargs):
        super().__init__(seq, **kwargs)
        self._log_initialized = False

    def add_env_config(self, prefix: str) -> None:
        """Convert environment variable such as "FROCKET_REDIS_HOST" to key "redis.host"."""
        for key, value in os.environ.items():
            value = value.strip()
            if key.startswith(prefix + '_') and len(value) > 0:
                clean_key = key.replace(prefix + '_', '').replace('_', '.').lower()
                self[clean_key] = value

    def _value_is_true(self, v: str = None) -> bool:
        return v and v.strip().lower() == 'true'

    def bool(self, key: str) -> bool:
        return self._value_is_true(self.get(key))

    def int(self, key) -> int:
        return int(self.get(key))

    def float(self, key) -> int:
        return float(self.get(key))

    def get_with_fallbacks(self, key: str, fallback_key: str, default: str = None) -> str:
        if key in self:
            return self[key]
        else:
            return self.get(fallback_key, default)

    def _get_for_service(self, key: str, service: str = None) -> str:
        if service:
            value = self.get_with_fallbacks(f"{service}.{key}", key, None)
        else:
            value = self.get(key, None)
        return value or None

    def aws_client_settings(self, service: str = None) -> dict:
        """
        Get the basic arguments for boto3.client(), to be unpacked into arguments by **aws_client_settings([service]).

        If service argument (e.g. 's3', 'lambda') is given, try first to read a value specific to that service,
        then fallback to reading a non service-specific value. If that one does not exist, boto would use its default
        mechanism of looking for env variables and configuration files.

        Having per-service settings if useful for (a) mocking the different services by separate tools (== endpoints),
        and (b) if using access keys rather than IAM roles, specifying service-specific keys.
        """
        aws_to_config_keys = {
            'region_name': 'aws.region',
            'endpoint_url': 'aws.endpoint.url',
            'aws_access_key_id': 'aws.access.key.id',
            'aws_secret_access_key': 'aws.secret.access.key'}
        settings = {aws_key: self._get_for_service(config_key, service)
                    for aws_key, config_key in aws_to_config_keys.items()}
        return settings

    def aws_config_dict(self, service: str = None) -> dict:
        """
        Returns values for initializing a boto Config object, to be unpacked by **aws_config_dict([service]).
        This is similar to the above method, except that these are values to init the optional config object with.
        """
        max_pool_connections = self._get_for_service('aws.max.pool.connections', service) or None
        connect_timeout = self._get_for_service('aws.connect.timeout', service) or None
        read_timeout = self._get_for_service('aws.read.timeout', service) or None
        max_retries = self._get_for_service('aws.retries.max.attempts', service) or None
        # For mocking purposes only!
        no_sign_requests = self._value_is_true(self._get_for_service('aws.no.signature', service))

        config = {
            'max_pool_connections': int(max_pool_connections) if max_pool_connections else None,
            'connect_timeout': int(connect_timeout) if connect_timeout else None,
            'read_timeout': int(read_timeout) if read_timeout else None,
            'retries': {'max_attempts': int(max_retries)} if max_retries else None,
            'signature_version': botocore.UNSIGNED if no_sign_requests else None
        }
        config = {k: v for k, v in config.items() if v is not None}
        return config

    @staticmethod
    def to_env_variable(key: str) -> str:
        """A helper for tests - enabling passing env variables to a sub-process with the current config value."""
        return ENV_CONFIG_PREFIX + '_' + key.replace('.', '_').upper()

    @property
    def loglevel(self) -> int:
        configured_level_name = self.get('log.level').upper()
        level = logging.getLevelName(configured_level_name)
        if not level:
            raise Exception(f"Uknown log level: {configured_level_name}")
        return level

    def _quiet_boto_logging(self):
        """botocore/boto3 can get pretty chatty, so only report warnings (e.g. connection pool pressure) and errors."""
        for package_prefix in ['botocore', 'boto3', 'urllib3']:
            boto3.set_stream_logger(name=package_prefix, level=logging.WARNING)

    def init_logging(self,
                     force_level: int = None,
                     force_console_output: bool = False) -> None:
        if not self._log_initialized:
            use_level = force_level or self.loglevel
            use_filename = None
            if not force_console_output:
                configured_filename = self.get('log.file')
                if configured_filename and len(configured_filename) > 0:
                    use_filename = configured_filename
            base_args = {
                'level': use_level,
                'format': self.get('log.format')
            }
            if use_filename:
                print(f"Logging to file {use_filename}")
                logging.basicConfig(filename=use_filename, **base_args)
            else:
                logging.basicConfig(stream=sys.stdout, **base_args)  # Please don't all go to stderr...

            self._quiet_boto_logging()
            self._log_initialized = True

    def init_lambda_logging(self):
        """
        Lambda Python runtime already init'ed a logger, so using basicConfig() as above would have no effect,
        and our precious log lines would have gone into /dev/oblivion. See https://stackoverflow.com/a/56579088
        """
        if not self._log_initialized:
            logging.getLogger().setLevel(self.loglevel)
            self._quiet_boto_logging()
            self._log_initialized = True


config = ConfigDict(DEFAULTS)
config.add_env_config(ENV_CONFIG_PREFIX)



"""
AsyncJobTracker object is handed by invoker_api to clients that launch a job in a non-blocking fashion.
It enables either periodic polling or blocking on updates. Updates are guaranteed to be atomic - that is,
there may be further updates, but the status you have in hand is consistent.
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import time
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from enum import auto
from queue import Queue, Empty
from typing import Optional, Dict, Generator
from frocket.common.serializable import AutoNamedEnum
from frocket.common.tasks.base import BaseJobResult, TaskStatus

logger = logging.getLogger(__name__)


class AsyncJobStage(AutoNamedEnum):
    STARTING = auto()
    RUNNING = auto()
    FINISHING = auto()
    DONE = auto()


@dataclass(frozen=True)
class AsyncJobStatus:
    stage: AsyncJobStage
    message: Optional[str] = None  # The job may set descriptive text for what it's doing
    result: Optional[BaseJobResult] = None  # Only available on stage=AsyncJobStage.DONE
    task_counters: Optional[Dict[TaskStatus, int]] = None


class JobTimeoutError(Exception):
    pass


class AsyncJobTracker(metaclass=ABCMeta):
    """The interface as known to clients"""

    @property
    @abstractmethod
    def status(self) -> AsyncJobStatus:
        """Get the latest status - as a consistent object which will not be mutated while using it"""
        pass

    @property
    @abstractmethod
    def elapsed_time(self) -> float:
        pass

    @property
    @abstractmethod
    def wait_time_remaining(self) -> Optional[float]:
        """
        If a tracker object was initialized with a timeout value by its creator (the invoker_api,
        based on configuration), then time remaining till timeout is known and can be returned.
        """
        pass

    @abstractmethod
    def wait(self, timeout: float = None) -> bool:
        """
        Blocking wait for updates with the given timeout, in seconds - but always capped to max wait time if set.
        By default, timeout is None - meaning wait up till max wait time (or indefinitely, in case it wasn't set).
        Assuming wait time is set, this is a good choice since no busy loop or semi-busy loop is needed.
        """
        pass

    def generator(self) -> Generator[AsyncJobStatus, None, None]:
        """
        Returns updates as they come for easier consumption, if blocking behavior is ok.
        This generator does not rely on any private attributes.
        """
        while True:
            update_available = self.wait()
            if not self.wait_time_remaining:
                raise JobTimeoutError()

            status_snapshot = self.status
            if status_snapshot.result:
                break

            if update_available:
                yield status_snapshot

        yield status_snapshot


class AsyncJobStatusUpdater(AsyncJobTracker):
    """
    Implementation of AsyncJobTracker, which is only created within invoker_api and updated by invoker/job code.

    The one curiousity here is the blocking wait() mechanism which is based on a Queue instance.
    How it works: the client's wait() call blocks on waiting for a queue item. If there's already one,
    it's immediately returned. Once consumed, the queue is empty again and a subsequent wait() will repeat
    the process. Typically, the queue should have either zero or a only single item - see _signal_update() below.
    """
    def __init__(self, max_wait: float = None):
        self._status: AsyncJobStatus = AsyncJobStatus(stage=AsyncJobStage.STARTING)
        self._update_queue = Queue()
        self._max_wait = max_wait
        self._start_time = time.time()

    @property
    def elapsed_time(self) -> float:
        return time.time() - self._start_time

    @property
    def wait_time_remaining(self) -> Optional[float]:
        assert self._max_wait
        remaining = self._max_wait - self.elapsed_time
        return remaining if remaining > 0 else 0

    @property
    def status(self) -> AsyncJobStatus:
        return self._status

    def _update_status(self, new_status: AsyncJobStatus) -> None:
        """Only signal an update if there was actually any change."""
        modified = self._status != new_status
        self._status = new_status
        if modified:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f"Updated async status from\n:{self._status} to:\n{new_status}")
            self._signal_update()
        pass

    def update(self, stage: AsyncJobStage = None, message: str = None, task_counters: Dict[TaskStatus, int] = None):
        # Asserts are used here as the invoker/job classes are internal to the invoker_api, and are expected to conform
        # to this class' requirements. If not, it's probably a bug.
        assert stage != AsyncJobStage.DONE  # To move to DONE stage, done() should be explicitly called
        assert self._status.stage != AsyncJobStage.DONE  # No more updates after DONE was called once
        stage = stage or self._status.stage
        task_counters = task_counters or self._status.task_counters
        # Automatically cleanup message when moving in stages
        message = message or (self._status.message if (stage == self._status.stage) else None)

        self._update_status(AsyncJobStatus(stage=stage, message=message,
                                           task_counters=task_counters))

    def done(self, result: BaseJobResult):
        self._update_status(AsyncJobStatus(stage=AsyncJobStage.DONE, result=result,
                                           task_counters=self._status.task_counters))

    def _signal_update(self):
        if self._update_queue.empty():
            # If the client *already* has an update waiting for it, no need to do anything - it will read the latest
            # state anyway when it gets to consume it (the queue item itself doesn't hold any information).
            # In case of more than single updater thread, there might momentarily be more than a single item.
            # However, this is not currently used in this way, and it seems that having multiple items would not
            # have any detrimental effect (i.e. break correctness) if it actually occurs in other edge cases.
            # TODO backlog to ensure a single item always, consider a lock here and re-test empty() within that lock.
            self._update_queue.put(object())

    def wait(self, timeout=None):
        assert timeout is None or timeout > 0
        try:
            should_block = True
            if self._max_wait:
                remaining = self.wait_time_remaining
                if remaining == 0:
                    # No more blocking wait - immediately return what's in the queue (or None)
                    should_block = False
                    timeout = None
                elif timeout:
                    timeout = min(timeout, remaining)
                else:
                    timeout = remaining

            self._update_queue.get(block=should_block, timeout=timeout)
            return True
        except Empty:
            return False



"""
Base classes for object serialization to JSON/de-serialization from JSON (which is needed both for communication
between invoker and workers, and for request/response handling in the API server), and a few data-related helpers.

I've chosen the package dataclasses-json as a basis, as it's playing nice with Python dataclasses and type hints,
instead of requiring special model classes. On top of that, a few extra capabilities and safeguards are introduced.
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
from collections import Counter
from dataclasses import dataclass, fields
from enum import Enum
from typing import List, Type, Dict, Any, Union, Optional
from datetime import datetime
import dataclasses_json
import marshmallow
from dataclasses_json import dataclass_json, LetterCase

# As suggested in dataclasses_json's docs, store datetime objects in ISO format to ensure timezone consistency (UTC)
dataclasses_json.global_config.encoders[datetime] = datetime.isoformat
dataclasses_json.global_config.decoders[datetime] = datetime.fromisoformat
dataclasses_json.global_config.mm_fields[datetime] = marshmallow.fields.DateTime(format='iso')


# For some reason, dataclasses-json package doesn't strictly validate the type of primitive fields,
# so here we are. The 'fallback_cls' below allows e.g. accepting an int where a float is expected.
def ensure_vtype(v, cls: Type, fallback_cls: Type = None):
    if not isinstance(v, cls) and \
            (not fallback_cls or not isinstance(v, fallback_cls)):
        raise ValueError(f"{v} is not a {cls.__name__}")
    return v


dataclasses_json.global_config.decoders[str] = lambda v: ensure_vtype(v, str)
dataclasses_json.global_config.decoders[int] = lambda v: ensure_vtype(v, int)
dataclasses_json.global_config.decoders[float] = lambda v: ensure_vtype(v, float, fallback_cls=int)
dataclasses_json.global_config.decoders[bool] = lambda v: ensure_vtype(v, bool)


def reducable(cls):
    """
    Decorator which marks a serializable class as 'reduceable' -  meaning a list of such objects can be aggregated
    to one by the reduce() class method.

    Reduceable classes must implement the *class method* _reduce_fields(), aceepting a list of similar-typed objects -
    see below. That method returns a dict with only the fields that were aggregated and their value, without any fields
    whose value should be common to all objects.
    """
    if not issubclass(cls, SerializableDataClass):
        raise Exception(f"{cls} is not serializable")
    cls.__reducable = True
    return cls


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(frozen=True)
class SerializableDataClass(dataclasses_json.DataClassJsonMixin):
    """
    Base class for all serializable classes. Since it's a dataclass, then construction, equality and hashing come out-
    of-the-box. Always immutable (frozen=True) to prevent coding errors and allow Python to safely generate __hash__().
    """

    def to_api_response_dict(self) -> dict:
        """
        Before the API server jsonifies are response, do some cleanups: remove None values and empty dicts,
        stringify enums (so a later json.dumps doesn't fail).
        TODO backlog support public/non-public mode? (including in nested objects)
        """
        def prepare_dict(src: dict, allowed_fields: List[str] = None) -> dict:
            res = {}
            for k, v in src.items():
                if v is None:
                    continue
                if isinstance(v, dict):
                    if len(v) == 0:
                        continue
                    res[k] = prepare_dict(v, allowed_fields=None)
                elif isinstance(v, Enum):
                    res[k] = str(v.value)
                else:
                    res[k] = v
            return res

        d = prepare_dict(self.to_dict())
        return d

    def shallowdict(self, include_none: bool = True):
        """dataclasses.asdict() would recurse into sub-dataclasses and make them dicts, this will keep them classes."""
        return {f.name: self.__getattribute__(f.name) for f in fields(self)
                if (include_none or self.__getattribute__(f.name) is not None)}

    @classmethod
    def reduce(cls, serializables: list) -> object:
        """
        Reduce a list of this class' type to one. The class must implement _reduce_fields() to perform the actual
        aggregation of relevant field, assisted by helper methods below. Fields expected to have a common non-reduced
        value are filled in by this method, and validated to be the same for all objects).
        """
        if not cls.is_reducable():
            raise Exception(f"{cls} is not marked as @reducable (and please implement _reduce_fields() for this class)")

        # Either all objects are None (and the result is thus None), or all must be of this class type
        if all([e is None for e in serializables]):
            return None
        else:
            assert all([type(e) is cls for e in serializables])

        base_fields = dataclasses.asdict(serializables[0])
        reduced_fields = cls._reduce_fields(serializables)  # Call the concrete class to handle all reducable fields
        # Ensure the non-reduced fields indeed have the same value for all objects in list
        common_fields = {k: v for k, v in base_fields.items() if k not in reduced_fields}
        for k, v in common_fields.items():
            assert all(getattr(e, k) == v for e in serializables[1:])

        all_fields = {**common_fields, **reduced_fields}
        # noinspection PyArgumentList
        return cls(**all_fields)

    @classmethod
    def is_reducable(cls) -> bool:
        return getattr(cls, '__reducable', False) is True

    @classmethod
    def _reduce_fields(cls, serializables: list) -> Dict[str, Any]:
        raise Exception(f"{cls.__name__} does not have its own _reduce_fields() implementation")

    @staticmethod
    def reduce_lists(lists: Union[List[list], List[None]]) -> Optional[list]:
        """A helper for subclasses implementing @reducable, for reducing a list field of other reducables."""
        if all(lst is None for lst in lists):
            return None
        else:
            list_size = len(lists[0])
            assert all([len(lst) == list_size for lst in lists])
            if list_size == 0:
                return []
            elemtype = type(lists[0][0])
            assert elemtype.is_reducable()
            assert all([type(lst[0]) == elemtype for lst in lists])

            reduced = [elemtype.reduce([e[i] for e in lists])
                       for i in range(list_size)]
            return reduced

    @staticmethod
    def reduce_counter_dicts(dicts: List[dict], top_count: int = None) -> dict:
        """A helper for reducing a list of dict-type fields, which are used as counters - and trimming the resutl len"""
        assert all(isinstance(d, dict) for d in dicts)
        reduced = Counter()
        for d in dicts:
            reduced.update(d)
        if top_count:
            reduced = dict(reduced.most_common(top_count))  # Sorts in any case
        return reduced


# TODO backlog decide on which of the two enum variants to settle - we want both serialization by name (so setting value
#  seems like the most robust way) and allowing extra attributes to be set per member. Python enums are tricksy!!!

class AutoNamedEnum(Enum):
    """
    Adapted from the Python enum documentation: when using auto(), set the enum value to the member name instead of
    the default auto-incrementing ordinal, to make its serialized form more human-friendly and play well with addition
    /re-ordering of enum members (but not renaming). However, since auto() is used, setting per-member attributes
    doesn't work.
    """

    # noinspection PyMethodParameters
    def _generate_next_value_(name, start, count, last_values):
        return name


class EnumSerializableByName(Enum):
    """Another go at it: this one doesn't set the member value to its name (__new__ doesn't know that name), and
    rather relies on setting a custom encoder/decoder below. Since it doesn't need auto(), other attributes can be set.
    This is all pretty frustrating."""
    def __new__(cls, *args):
        obj = object.__new__(cls)
        obj._value_ = len(cls.__members__) + 1
        return obj

    # noinspection PyUnusedLocal
    def __init__(self, *args):
        cls = self.__class__
        if cls not in dataclasses_json.global_config.encoders:
            dataclasses_json.global_config.encoders[cls] = lambda e: e.name
            dataclasses_json.global_config.decoders[cls] = lambda n: self.__class__[n]


"""
Envelope - holds a SerializableDataClass and its type for later de-serializing the correct concrete type, 
similar to https://awslabs.github.io/aws-lambda-powertools-python/utilities/parser/#envelopes

Currently, only classes decorated with @enveloped (see below) are allowed, to prevent misuse.
"""

enveloped_classes = []


def enveloped(cls):
    if not issubclass(cls, SerializableDataClass):
        raise Exception(f"{cls} is not serializable")
    enveloped_classes.append(cls)
    return cls


@dataclass(frozen=True)
class Envelope(SerializableDataClass):
    content_cls: str
    content_dict: dict

    @classmethod
    def seal(cls, obj: SerializableDataClass) -> SerializableDataClass:
        if type(obj) not in enveloped_classes:
            raise Exception(f"Type {type(obj)} is not marked as @enveloped")
        return cls(content_cls=type(obj).__name__, content_dict=obj.to_dict())

    @classmethod
    def seal_to_json(cls, obj: SerializableDataClass) -> str:
        return cls.seal(obj).to_json()

    def open(self, expected_superclass: Type[SerializableDataClass] = None) -> SerializableDataClass:
        content_class = next((cls for cls in enveloped_classes if cls.__name__ == self.content_cls), None)
        if not content_class:
            raise Exception(f"Type {self.content_cls} is not marked as @enveloped")
        elif expected_superclass and not issubclass(content_class, expected_superclass):
            raise Exception(f"Type {content_class} is not a subclass of {expected_superclass}")

        return content_class.from_dict(self.content_dict)

    @classmethod
    def open_from_json(cls, json_string: str,
                       expected_superclass: Type[SerializableDataClass] = None) -> SerializableDataClass:
        envelope = Envelope.from_json(json_string)
        return envelope.open(expected_superclass)



"""
A worker that gets its tasks by a blocking dequeue from the datastore. Doesn't get any simpler -
but is easily scalable, and requires no load balancer or orchestrator (except for the queue's atomic guarantees).

TODO backlog having a cache-friendly task assignment would require more work, if it makes sense to do.
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from frocket.common.metrics import MetricsBag, WorkerStartupLabel, ComponentLabel
from frocket.common.tasks.base import BaseTaskRequest
from frocket.datastore.registered_datastores import get_datastore
from frocket.worker.impl.generic_env_metrics import GenericEnvMetricsProvider
from frocket.worker.runners.base_task_runner import BaseTaskRunner, TaskRunnerContext
from frocket.common.config import config
from frocket.worker.runners.registered_runners import REGISTERED_RUNNERS

config.init_logging()
logger = logging.getLogger(__name__)
datastore = get_datastore()


def handle(req: BaseTaskRequest) -> None:
    metrics = MetricsBag(component=ComponentLabel.WORKER,
                         env_metrics_provider=GenericEnvMetricsProvider())
    metrics.set_label_enum(WorkerStartupLabel.WARM)  # Always warm this worker is, uhmmhmmhmmhmm

    runner_class = REGISTERED_RUNNERS[type(req)]
    runner = runner_class(req, TaskRunnerContext(metrics))
    result = runner.run()
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug(result.to_json())


def main_loop():
    # TODO backlog currently workers that encounter an unexpected data format will crash rather than continuing to
    #  consume and (probably) fail. This has a pro (outdated worker versions fail fast), but of course also cons -
    #  consider the desired/configurable behavior (e.g. crash after N unexpected errors?)
    try:
        while True:
            logger.info('Waiting for work...')
            req: BaseTaskRequest = datastore.dequeue()
            if req:
                logger.info(f"Got request: {req}")

                should_run, reject_reason = BaseTaskRunner.should_run(req)
                if should_run:
                    handle(req)
                else:
                    logger.warning(f"Request rejected: {reject_reason}")
    except KeyboardInterrupt:
        logger.info('Bye')


main_loop()



"""
Query job's task classes
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from dataclasses import dataclass
from enum import auto
from typing import Optional, List, Dict, Union, cast
import inflection
from frocket.common.dataset import DatasetInfo, DatasetPartId
from frocket.common.serializable import AutoNamedEnum, enveloped, SerializableDataClass, reducable
from frocket.common.tasks.base import BaseTaskRequest, BaseTaskResult, BaseJobResult


class PartSelectionMode(AutoNamedEnum):
    """Whether the invoker sets the task_index or the worker selects it from available tasks in the datastore."""
    SET_BY_INVOKER = auto()
    SELECTED_BY_WORKER = auto()


@enveloped
@dataclass(frozen=True)
class QueryTaskRequest(BaseTaskRequest):
    dataset: DatasetInfo
    # String columns to load as Pandas categoricals, as performance optimization. These columns are detected during
    # dataset registration. Not needed for columns already of categorical type in files saved by Pandas.
    load_as_categoricals: Optional[List[str]]
    mode: PartSelectionMode
    # If (and only if) mode=SET_BY_INVOKER, the invoker also sets the dataset part index to query
    # Note that task_index not necessarily equals part ID
    invoker_set_part: Optional[DatasetPartId]
    used_columns: List[str]  # Which columns to actually load (as optimization), as analyzed by QueryValidator.
    query: dict


class AggregationType(AutoNamedEnum):
    # noinspection PyUnusedLocal
    def __init__(self, *args):
        if not hasattr(self.__class__, '_camels'):
            self.__class__._camels = {}

        self.camelized = inflection.camelize(self.name.lower(), uppercase_first_letter=False)
        self.__class__._camels[self.camelized] = self
        self.value_is_dict = self.name.endswith("_PER_VALUE")

    COUNT = auto()
    COUNT_PER_VALUE = auto()
    GROUPS_PER_VALUE = auto()
    SUM_PER_VALUE = auto()
    MEAN_PER_VALUE = auto()

    @classmethod
    def from_camelcase(cls, camelcase_name: str) -> AutoNamedEnum:
        return cls._camels[camelcase_name]


AggrValue = Union[int, float]
AggrValueMap = Dict[str, AggrValue]


@reducable
@dataclass(frozen=True)
class AggregationResult(SerializableDataClass):
    column: str
    type: str
    # For some aggregation types ('count') the value is a single number. In others (the '<X>perValue' ones), value is
    # a dict of column value->aggregated number
    value: Optional[Union[AggrValue, AggrValueMap]]
    top: Optional[int]  # Relevant for values of type dict
    name: Optional[str]  # Only set if the user has set a custom name for this aggregation

    @classmethod
    def _reduce_fields(cls, serializables):
        """See: SerializableDataClass."""
        all_values = [e.value for e in cast(List[AggregationResult], serializables)]
        # Reduce either a primitive values or a dicts of counters
        if isinstance(all_values[0], dict):
            reduced_value = cls.reduce_counter_dicts(all_values, top_count=cast(cls, serializables[0]).top)
        else:
            reduced_value = sum(all_values)
        return {'value': reduced_value}


@reducable
@dataclass(frozen=True)
class QueryConditionsResult(SerializableDataClass):
    matching_groups: int  # e.g. user ID
    matching_group_rows: int  # All rows of the matching groups, whether that row matches a condition or not
    aggregations: Optional[List[AggregationResult]]

    @classmethod
    def _reduce_fields(cls, serializables):
        results = cast(List[cls], serializables)
        return {'matching_groups': sum([e.matching_groups for e in results]),
                'matching_group_rows': sum([e.matching_group_rows for e in results]),
                'aggregations': cls.reduce_lists([e.aggregations for e in results])}


@reducable
@dataclass(frozen=True)
class FunnelResult(SerializableDataClass):
    sequence: List[QueryConditionsResult]
    end_aggregations: Optional[List[AggregationResult]]

    @classmethod
    def _reduce_fields(cls, serializables):
        funnel_results = cast(List[cls], serializables)
        return {'sequence': cls.reduce_lists([e.sequence for e in funnel_results]),
                'end_aggregations': cls.reduce_lists([e.end_aggregations for e in funnel_results])}


@reducable
@dataclass(frozen=True)
class QueryResult(SerializableDataClass):
    query: QueryConditionsResult
    funnel: Optional[FunnelResult]

    @classmethod
    def _reduce_fields(cls, serializables):
        query_results = cast(List[cls], serializables)
        return {'query': QueryConditionsResult.reduce([e.query for e in query_results]),
                'funnel': FunnelResult.reduce([e.funnel for e in query_results])}


@enveloped
@dataclass(frozen=True)
class QueryTaskResult(BaseTaskResult):
    query_result: Optional[QueryResult]  # Not set if query failed (when success=False)


@dataclass(frozen=True)
class QueryJobResult(BaseJobResult):
    query: Optional[QueryConditionsResult]
    funnel: Optional[FunnelResult]



"""
Classic Map/Reduce-style dataset re-partitioning utility.

Given a set of input files in Parquet format and a numeric column name, shuffle all rows into the given no. of parts
(a set of numbered files or directories). The appropriate part ID for each row is set by simple 'value % num_parts' over
the given column value.

This is a mandatory requirement for using datasets in Funnel Rocket, since it ensures that all rows of each specific
user/group ID (or any other identifier you wish to partition by) are found in a single file only, thus no shuffle
is necessary at query time.

This utility is not part of Funnel Rocket package itself. Rather, it can assist in creating datasets of limited size -
since it only runs on a single machine. It does utilize multiple CPUs, so running on a beefy VM is an option for ad-hoc
larger jobs. For a robust & scalable solution, using PySpark or the like is advised.

TODO consider a distributed re-partitioning feature in Funnel Rocket (atm it's a read-only tool)
"""
import pyarrow as pa
import pyarrow.parquet as pq
import pyarrow.dataset as ds
import glob
import argparse
import sys
import pathlib
import math
import shutil
from multiprocessing import Pool, cpu_count

MAX_PART_NUMBER_DIGITS = 5


def get_part_path(outdir_path, i):
    """Helper for naming parts, with part number i padded by zeros (e.g. 'map/part-00037'), which makes filenames
    naturally sorted."""
    padded_part_number = str(i).zfill(MAX_PART_NUMBER_DIGITS)
    part_part = outdir_path / 'part-{}'.format(padded_part_number)
    return part_part


def get_worker_pool_size(num_files):
    """Returns how many worker processes to use."""
    available_cpus = max(cpu_count() - 1, 1)  # Leave one CPU alone, if there's more than one
    num_workers = min(available_cpus, num_files)  # No need for more workers than the actual no. of files to process
    return num_workers


def map_worker_task(task_id, input_file_path, outdir_path, partition_by, num_parts):
    """A single mapper task: read one input file and map its contents into N parts by the partition_by column."""
    table = pq.read_table(input_file_path, use_threads=False)  # Single-threaded, as we're using all CPUs anyway

    # id_mod_series will hold the appropriate part number for each row in the table
    id_series = table.column(partition_by).to_pandas()
    id_mod_series = id_series % num_parts

    # Iterate over part numbers. In each iteration, filter only the rows relevant for current part and save them
    for part_number in range(num_parts):
        # Create a boolean "mask" column (True only for rows where partition_by field maps to to current part no.)
        mask = (id_mod_series == part_number).values  # Get NumPy boolean array
        masked_table = table.filter(mask)

        # The file will be placed in the part folder, but it needs a unique name
        # (it could also be the value of i, but here we use the original file name)
        output_file_path = get_part_path(outdir_path, part_number) / 'from-{}.parquet'.format(input_file_path.stem)
        pq.write_table(masked_table, output_file_path)

        if part_number and part_number % 100 == 0:
            print(f"Task no. {task_id} for input file {input_file_path}: written {part_number} files so far... ")


def map_files(args):
    """Run map tasks for all input files."""
    files_list = glob.glob(args.input, recursive=True)
    if not files_list:
        sys.exit(f"No input files matching pattern '{args.input}'")

    map_dir_path = pathlib.Path(args.mapdir)
    existing_files = list(map_dir_path.glob('*'))
    if len(existing_files) > 0:
        if args.force:
            print(f"Recreating map output dir '{map_dir_path}'")
            shutil.rmtree(map_dir_path)
        else:
            sys.exit(f"Map output dir '{map_dir_path}' is not empty (consider using the --force)")

    map_dir_path.mkdir(parents=True, exist_ok=True)

    # Create sub-directory for each part - these directories are needed by all tasks that will emit their part,
    # so create them centrally here (and not inside a task)
    for i in range(args.parts):
        part_path = get_part_path(map_dir_path, i)
        if not part_path.exists():
            part_path.mkdir()

    pool_size = get_worker_pool_size(len(files_list))
    print(f"Input files found: {len(files_list)}, map output dir: {map_dir_path}, target partitions: {args.parts}, "
          f"CPUs: {cpu_count()}, pool size: {pool_size}")

    # Prepare a task for each input file
    pool = Pool(pool_size)
    tasks = []
    for task_id, input_file in enumerate(files_list):
        task_args = (task_id,
                     pathlib.Path(input_file),
                     map_dir_path,
                     args.partition_by,
                     args.parts)
        tasks.append(pool.apply_async(map_worker_task, task_args))

    # Actually run all tasks to completion
    [task.get() for task in tasks]

    # Ensure that the expected number of output files are found
    expected_outfiles_count = len(files_list) * args.parts
    outfiles_found = map_dir_path.glob('part-*/*.parquet')
    actual_outfiles_count = len(list(outfiles_found))

    if actual_outfiles_count != expected_outfiles_count:
        sys.exit(f"Expected {expected_outfiles_count} map output files, but found {actual_outfiles_count}")

    print(f"Map stage done, total of {actual_outfiles_count} files created")


def reduce_worker_task(part_map_path, part_reduce_path, rows_per_file, sort_by):
    """
    A single reduce task: load all intermediate map output files for its assigned part number,
    merging all rows into one partition.

    TODO to limit memory usage in real-world jobs, read files iteratively rather than all at once
    """
    dataset = ds.dataset(part_map_path, format="parquet")
    table = dataset.to_table()

    if sort_by:
        # Go through Pandas for sorting (this is heavy and very memory-hungry)
        df = table.to_pandas()
        df.sort_values(by=sort_by, inplace=True, ignore_index=True)
        # noinspection PyArgumentList
        table = pa.Table.from_pandas(df, schema=table.schema)

    if rows_per_file:
        num_files = math.ceil(table.num_rows / rows_per_file)
        for slice_number in range(num_files):
            if slice_number == (num_files - 1):  # Last slice (or only slice)
                length = None  # No lenght limit (see below)
            else:
                length = rows_per_file

            offset = slice_number * rows_per_file
            curr_slice = table.slice(offset, length)
            output_file_path = part_reduce_path / 'slice-{}.parquet'.format(slice_number)
            pq.write_table(curr_slice, output_file_path)
    else:
        output_file_path = str(part_reduce_path) + '.parquet'
        pq.write_table(table, output_file_path)


def reduce_files(args):
    """Run a reduce task for each part number in the target no. of partitions."""
    map_dir_path = pathlib.Path(args.mapdir)
    if not map_dir_path.is_dir():
        sys.exit(f"Map output dir '{map_dir_path}' does not exists, did you run the map stage?")

    num_part_dirs = len(list(map_dir_path.glob('part-*/')))
    if num_part_dirs != args.parts:
        sys.exit(f"Expected {args.parts} sub-dirs under map dir, but found {num_part_dirs}")

    reduce_dir_path = pathlib.Path(args.reducedir)
    existing_files = list(reduce_dir_path.glob('*'))
    if len(existing_files) > 0:
        if args.force:
            print(f"Recreating reduce output dir '{reduce_dir_path}'")
            shutil.rmtree(reduce_dir_path)
        else:
            sys.exit(f"Reduce output dir '{reduce_dir_path}' is not empty (consider using the --force)")

    reduce_dir_path.mkdir(parents=True, exist_ok=True)

    # Create sub-directory per output part - if allowing multi-file partitions
    if args.maxrows:
        for i in range(args.parts):
            part_path = get_part_path(reduce_dir_path, i)
            if not part_path.exists():
                part_path.mkdir()

    # Collect the input paths: we expect an existing sub-dir generated in the map stage, per each part number
    paths = []
    for i in range(args.parts):
        part_path = get_part_path(map_dir_path, i)
        if not part_path.exists():
            sys.exit(f"Part path '{part_path}' does not exist!")
        paths.append(part_path)

    # Prepate a task per part
    pool = Pool(get_worker_pool_size(len(paths)))
    tasks = []
    for task_number, part_path in enumerate(paths):
        part_reduce_path = get_part_path(reduce_dir_path, task_number)
        task_args = (part_path, part_reduce_path, args.maxrows, args.sort)
        tasks.append(pool.apply_async(reduce_worker_task, task_args))

    # Run all tasks to completion
    [task.get() for task in tasks]
    print(f"Reduce stage done, output parts are in directory: {reduce_dir_path}")


if __name__ == '__main__':  # Don't run in worker processes
    DEFAULT_NUM_PARTS = 16
    DEFAULT_INPUT_FILES = './data/**/*.parquet'
    DEFAULT_MAP_DIR = './map'
    DEFAULT_REDUCE_DIR = './reduce'

    parser = argparse.ArgumentParser(
        description='Partition Parquet files into N parts in two stages: map & reduce',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('partition_by', type=str,
                        help='Numeric column to partition by (e.g. userId, experimentId, ...)')
    parser.add_argument('--action', type=str,
                        choices=['all', 'map', 'reduce'], default='all',
                        help='Whether to run just the map or reduce stages')
    parser.add_argument('--input', type=str,
                        default=DEFAULT_INPUT_FILES,
                        help='A glob-style pattern of input files to map')
    parser.add_argument('--mapdir', type=str,
                        default=DEFAULT_MAP_DIR,
                        help='Root directory for map output')
    parser.add_argument('--reducedir', type=str,
                        default=DEFAULT_REDUCE_DIR,
                        help='Root directory for reduced files')
    parser.add_argument('--parts', type=int,
                        default=DEFAULT_NUM_PARTS,
                        help='number of partitions to emit')
    parser.add_argument('--force', default=False,
                        action='store_true',
                        help='Whether to overwrite existing files')
    parser.add_argument('--maxrows', type=int,
                        help='Optionally split the output files by a maximum N rows per file, to limit file size. '
                             'In this mode each part is a directory with one or more files, rather than a single file.')
    parser.add_argument('--sort', type=str, nargs='+',
                        help='Optional space-separated list of column names to sort by in the reduce stage. '
                        'By default, output files are not sorted. NOTE: This is memory- and compute-intensive.')

    args = parser.parse_args()
    if args.action in ['map', 'all']:
        map_files(args)
    if args.action in ['reduce', 'all']:
        reduce_files(args)



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import time
from contextlib import contextmanager
from typing import Type, List, cast, Optional
from frocket.common.dataset import DatasetInfo, DatasetPartsInfo, DatasetShortSchema
from frocket.common.helpers.utils import timestamped_uuid
from frocket.common.metrics import MetricsBag, ComponentLabel
from frocket.common.tasks.base import BaseTaskRequest, BaseTaskResult, BaseJobResult, TaskStatus, TaskAttemptsInfo, \
    JobStatus, TaskAttemptId, JobStats
from frocket.common.tasks.query import QueryTaskResult, QueryTaskRequest, QueryJobResult
from frocket.common.tasks.registration import RegistrationTaskRequest, RegistrationTaskResult, RegistrationJobResult, \
    DatasetValidationMode, REGISTER_DEFAULT_FILENAME_PATTERN, RegisterArgs
from frocket.common.validation.query_validator import QueryValidator
from frocket.datastore.registered_datastores import get_datastore
from frocket.invoker.base_invoker import BaseInvoker
from frocket.invoker.jobs.job import Job
from frocket.invoker.jobs.query_job import QueryJob
from frocket.invoker.jobs.registration_job import RegistrationJob
from frocket.worker.impl.generic_env_metrics import GenericEnvMetricsProvider
from frocket.worker.runners.base_task_runner import TaskRunnerContext
from frocket.worker.runners.part_loader import PartLoader
from frocket.worker.runners.registered_runners import REGISTERED_RUNNERS
from tests.utils.dataset_utils import DEFAULT_GROUP_COLUMN, DEFAULT_TIMESTAMP_COLUMN, TestDatasetInfo
# noinspection PyUnresolvedReferences
from tests.utils.redis_fixture import init_test_redis_settings


def simple_run_task(req: BaseTaskRequest,
                    expected_resultcls: Type[BaseTaskResult],
                    private_part_loader: PartLoader = None,
                    preflight_duration_ms: int = None) -> BaseTaskResult:
    ctx = TaskRunnerContext(metrics=MetricsBag(component=ComponentLabel.WORKER,
                                               env_metrics_provider=GenericEnvMetricsProvider()),
                            private_part_loader=private_part_loader,
                            preflight_duration_ms=preflight_duration_ms)
    task_runner = REGISTERED_RUNNERS[type(req)](req, ctx)
    result = task_runner.run()
    assert type(result) == expected_resultcls
    return result


class InprocessInvokerAndWorker:
    def __init__(self, job: Job,
                 taskreq_cls: Type[BaseTaskRequest],
                 taskres_cls: Type[BaseTaskResult],
                 jobres_cls: Type[BaseJobResult]):
        self._job = job
        self._taskreq_cls = taskreq_cls
        self._taskres_cls = taskres_cls
        self._jobres_cls = jobres_cls
        self._queue_name = timestamped_uuid('queue-')
        self.num_tasks: Optional[int] = None
        self._job_status: Optional[JobStatus] = None
        self._all_task_results: Optional[List[BaseTaskResult]] = None
        self._latest_task_results: Optional[List[BaseTaskResult]] = None
        self._final_job_status: Optional[JobStatus] = None

    def prerun(self):
        self._job.prerun()

    def enqueue_tasks(self) -> List[BaseTaskRequest]:
        requests = cast(List[self._taskreq_cls], self._job.build_tasks())
        self.num_tasks = len(requests)
        parts_to_publish = self._job.dataset_parts_to_publish()
        if parts_to_publish:
            get_datastore().publish_for_worker_selection(reqid=self._job.request_id,
                                                         attempt_round=0,
                                                         parts=parts_to_publish)
        get_datastore().enqueue(requests, queue=self._queue_name)
        return requests

    def run_tasks(self):
        return self._run_tasks(self.num_tasks)

    def _run_tasks(self, expected_task_count: int):
        for i in range(expected_task_count):
            req = get_datastore().dequeue(queue=self._queue_name, timeout=1)
            assert req
            result = simple_run_task(req, self._taskres_cls)
            assert result.status == TaskStatus.ENDED_SUCCESS
        assert get_datastore().dequeue(queue=self._queue_name, timeout=0) is None  # Queue depleted

        all_attempts = get_datastore().task_results(self._job.request_id)
        assert len(all_attempts) >= self.num_tasks
        self._all_task_results = cast(List[self._taskres_cls], list(all_attempts.values()))
        assert all([res.status == TaskStatus.ENDED_SUCCESS for res in self._all_task_results])

        tasks_status = get_datastore().tasks_status(self._job.request_id)
        attempts_status = [TaskAttemptsInfo(i) for i in range(self.num_tasks)]
        for at, update in tasks_status.items():
            attempts_status[at.task_index].add(at, update)
        assert all([tat.latest_update.status == TaskStatus.ENDED_SUCCESS for tat in attempts_status])

        self._job_status = JobStatus(success=True, error_message=None, attempts_status=attempts_status)
        self._latest_task_results = BaseInvoker.latest_results_only(self._job_status, all_attempts)
        return self._all_task_results

    def retry(self, task_index) -> List[BaseTaskResult]:
        attempt_no = get_datastore().increment_attempt(self._job.request_id, task_index)
        attempt_id = TaskAttemptId(task_index, attempt_no)

        req = self._job.build_retry_task(attempt_no, task_index)
        get_datastore().enqueue([req], queue=self._queue_name)
        results = self._run_tasks(expected_task_count=1)
        return results

    def complete(self, assert_success: bool = True) -> JobStatus:
        self._final_job_status = self._job.complete(self._job_status, latest_task_results=self._latest_task_results)
        if assert_success:
            assert self._final_job_status.success and not self._final_job_status.error_message
        return self._final_job_status

    def build_result(self, assert_success: bool = True) -> BaseJobResult:
        base_attributes = BaseJobResult(
            request_id=self._job.request_id,
            success=self._final_job_status.success,
            error_message=self._final_job_status.error_message,
            stats=JobStats()). \
            shallowdict(include_none=True)

        job_result = self._job.build_result(base_attributes, self._final_job_status, self._latest_task_results)
        assert type(job_result) is self._jobres_cls
        if assert_success:
            assert job_result.success
        return job_result

    def run_all_stages(self) -> BaseJobResult:
        self.prerun()
        self.enqueue_tasks()
        self.run_tasks()
        self.complete()
        return self.build_result()

    def cleanup(self):
        get_datastore().cleanup_request_data(self._job.request_id)


def build_registration_job(basepath: str,
                           mode: DatasetValidationMode,
                           group_id_column: str = DEFAULT_GROUP_COLUMN,
                           pattern: str = REGISTER_DEFAULT_FILENAME_PATTERN,
                           uniques: bool = True) -> RegistrationJob:
    args = RegisterArgs(
        name=f"test-{basepath}-{time.time()}",
        basepath=basepath,
        group_id_column=group_id_column,
        timestamp_column=DEFAULT_TIMESTAMP_COLUMN,
        pattern=pattern,
        validation_mode=mode,
        validate_uniques=uniques)
    job = RegistrationJob(args=args)
    job.request_id = timestamped_uuid('test-reg-')
    return job


@contextmanager
def registration_job(test_ds: TestDatasetInfo,
                     mode: DatasetValidationMode,
                     group_id_column: str = DEFAULT_GROUP_COLUMN,
                     pattern: str = REGISTER_DEFAULT_FILENAME_PATTERN,
                     uniques: bool = True) -> RegistrationJob:
    job = build_registration_job(test_ds.basepath, mode, group_id_column, pattern, uniques)
    try:
        yield job
    finally:
        if job.dataset:
            get_datastore().remove_dataset_info(job.dataset.id.name)


def build_query_job(query: dict,
                    ds: DatasetInfo,
                    parts: DatasetPartsInfo,
                    short_schema: DatasetShortSchema,
                    worker_can_select_part: bool = None) -> QueryJob:
    validation_result = QueryValidator(query, ds, short_schema).expand_and_validate()
    job = QueryJob(ds, parts, short_schema,
                   query=validation_result.expanded_query,
                   used_columns=validation_result.used_columns,
                   worker_can_select_part=worker_can_select_part)
    job.request_id = timestamped_uuid('test-query-')
    return job


@contextmanager
def registration_job_invoker(job: RegistrationJob) -> InprocessInvokerAndWorker:
    invoker_runner = InprocessInvokerAndWorker(
        job, RegistrationTaskRequest, RegistrationTaskResult, RegistrationJobResult)
    try:
        yield invoker_runner
    finally:
        invoker_runner.cleanup()


@contextmanager
def query_job_invoker(job: QueryJob) -> InprocessInvokerAndWorker:
    invoker_runner = InprocessInvokerAndWorker(
        job, QueryTaskRequest, QueryTaskResult, QueryJobResult)
    try:
        yield invoker_runner
    finally:
        invoker_runner.cleanup()



"""
Load and cache parts (data files).
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import time
import os
from pathlib import Path
from typing import List, Dict, Optional, Set, NamedTuple, Union
from pandas import DataFrame
import pyarrow.parquet
from frocket.common.config import config
from frocket.common.helpers.storage import storage_handler_for
from frocket.common.helpers.utils import memoize
from frocket.common.metrics import MetricName, LoadFromLabel, MetricsBag
from frocket.common.dataset import DatasetPartId, DatasetId

logger = logging.getLogger(__name__)


# Just a little typed nicety over tuples which PyArrow accepts as predicate pushdown filters
class FilterPredicate(NamedTuple):
    column: str
    op: str
    value: Union[str, int, float, bool]


class CacheEntry:
    local_path: str
    size_mb: float
    last_used: float


class PartLoader:
    _cache: Dict[DatasetPartId, CacheEntry] = None  # DatasetPartId is a dataclass with proper hash & equality
    _disk_cache_max_size: float = None

    def __init__(self):
        self._setup()

    # Support re-initialization and overriding the configured size, for testing
    def _setup(self, disk_cache_max_size: float = None):
        if self._cache:
            for entry in self._cache.values():
                os.remove(entry.local_path)
        self._cache = {}
        self._disk_cache_max_size = disk_cache_max_size if disk_cache_max_size is not None \
            else config.float('worker.disk.cache.size.mb')

    @property
    def cache_current_size_mb(self) -> float:
        return sum(entry.size_mb for entry in self._cache.values())

    @property
    def cache_len(self) -> int:
        return len(self._cache)

    def _prune_cache(self) -> None:
        curr_size_mb = self.cache_current_size_mb
        while curr_size_mb > 0 and curr_size_mb > self._disk_cache_max_size:
            logger.info(f"Current cache size is {curr_size_mb}mb, more than the configured "
                        f"{self._disk_cache_max_size}mb")
            lru_key = min(self._cache, key=lambda k: self._cache[k].last_used)
            lru_entry = self._cache[lru_key]
            logger.info(f"Deleting LRU entry of dataset: {lru_key.dataset_id.name} "
                        f"source path: {lru_key.path}, "
                        f"last used {time.time() - lru_entry.last_used:.1f} seconds ago")
            try:
                os.remove(lru_entry.local_path)
            except OSError:
                logger.exception('Failed to delete file!')  # TODO backlog consider disabling any further caching
            del self._cache[lru_key]
            curr_size_mb = self.cache_current_size_mb

    def load_dataframe(self,
                       file_id: DatasetPartId,
                       metrics: MetricsBag,
                       needed_columns: List[str] = None,
                       filters: List[FilterPredicate] = None,
                       load_as_categoricals: List[str] = None) -> DataFrame:
        self._prune_cache()
        loaded_from: Optional[LoadFromLabel] = LoadFromLabel.SOURCE
        handler = storage_handler_for(file_id.path)
        is_source_remote = handler.remote

        local_path = None
        if not is_source_remote:
            local_path = file_id.path  # No caching for local files
        else:
            if file_id in self._cache:
                local_path = self._cache[file_id].local_path
                loaded_from = LoadFromLabel.DISK_CACHE
                self._cache[file_id].last_used = time.time()
                logger.info("File is locally cached, yay")

        if not local_path:
            with metrics.measure(MetricName.TASK_DOWNLOAD_SECONDS):
                local_path = str(handler.get_local_path(file_id.path))  # Download to a local temp file

            entry = CacheEntry()
            entry.local_path = local_path
            entry.size_mb = Path(local_path).stat().st_size / 1024 ** 2
            entry.last_used = time.time()
            self._cache[file_id] = entry

        with metrics.measure(MetricName.TASK_LOAD_FILE_SECONDS):
            # Using PyArrow directly (rather than wrapped through Pandas) allows specifying column names to explicitly
            # load as 'dictionary' type, which then translates to categoricals in Pandas.
            # If the file was created with Pandas, categorical columns are loaded back as such - but we go beyond
            # that to detect 'potential categorical' string columns and load them as such.
            # Except for the memory usage saving, there is a performance gain here if the Parquet file already has a
            # dictionary for the column. Otherwise, PyArrow will create one - but without a performance gain.
            df = pyarrow.parquet.read_table(local_path,
                                            columns=needed_columns,
                                            filters=filters,
                                            read_dictionary=load_as_categoricals).to_pandas()

        metrics.set_label_enum(loaded_from)
        return df

    def get_cached_candidates(self, dataset_id: DatasetId) -> Optional[Set[DatasetPartId]]:
        """Do we have cached parts for this DatasetId, that can be used to self-select parts?"""
        logger.debug(f"Looking for cached candidates matching: {dataset_id}")
        candidates = None
        if self._cache:
            candidates = {part_id for part_id in self._cache.keys() if part_id.dataset_id == dataset_id}

        logger.debug(f"Found candidates: {candidates}")
        return candidates if (candidates and len(candidates) > 0) else None


@memoize
def shared_part_loader() -> PartLoader:
    """This is used by default, but can be overriden in tests."""
    return PartLoader()



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json

from frocket.common.validation.path_visitor import PathVisitor

root = {
    "robot": False,
    "person": {
        "name": {
            "first": "elad", "last": "roz"
        },
        "primitives": [1, 2,
                       {"a": 3},
                       {"b": {"c": [4, 5, {"ping": "pong1"}]}},
                       {"b": {"c": [6, 7, {"ping": "pong2"}]}},
                       ],
        "nicks": [
            {"short": "e"},
            {"short": "roz"},
            {"short": None},
            {"long": {"heya": [123]}}
        ]
    }
}

original_root_json = json.dumps(root, sort_keys=True)


def root_clone():
    return json.loads(original_root_json)


def dict_equals_root(dct: dict):
    return json.dumps(dct, sort_keys=True) == original_root_json


def test_empty_root():
    assert PathVisitor(root={}, path="a.b.c").list() == []


def test_empty_path():
    assert PathVisitor(root=root, path="").list() == []
    assert PathVisitor(root=root, path="    ").list() == []


def test_robot():
    assert PathVisitor(root=root, path="robot").list() == [False]
    assert PathVisitor(root=root, path=" robot ").list() == [False]
    assert PathVisitor(root=root, path="robot.").list() == []


def test_toplevel_miss():
    assert PathVisitor(root=root, path="pers").list() == []
    assert PathVisitor(root=root, path=".pers.").list() == []


def test_person():
    assert PathVisitor(root=root, path="person").list() == [root['person']]
    assert PathVisitor(root=root, path="person.name").list() == [root['person']['name']]
    assert PathVisitor(root=root, path="person.name.first").list() == [root['person']['name']['first']]
    assert PathVisitor(root=root, path="person.name.last").list() == [root['person']['name']['last']]
    assert PathVisitor(root=root, path="person.name.last.").list() == []
    assert PathVisitor(root=root, path="person.name.lastx").list() == []
    assert PathVisitor(root=root, path="person.name.last.really.now").list() == []


def test_primitives():
    assert PathVisitor(root=root, path="person.primitives").list() == \
           root["person"]["primitives"]
    assert PathVisitor(root=root, path="person.primitives", list_to_items=False).list() == \
           [root["person"]["primitives"]]

    assert PathVisitor(root=root, path="person.primitives.").list() == []
    assert PathVisitor(root=root, path="person.primitives.", list_to_items=False).list() == []

    assert PathVisitor(root=root, path="person.primitives.a").list() == \
           [root["person"]["primitives"][2]["a"]]

    assert PathVisitor(root=root, path="person.primitives.b").list() == \
           [root["person"]["primitives"][3]["b"], root["person"]["primitives"][4]["b"]]

    # Should return a list of all elements in the matching two lists
    # noinspection PyUnresolvedReferences
    assert PathVisitor(root=root, path="person.primitives.b.c").list() == \
           [*root["person"]["primitives"][3]["b"]["c"], *root["person"]["primitives"][4]["b"]["c"]]

    # Should return a list with the matching two lists as they are
    # noinspection PyUnresolvedReferences
    assert PathVisitor(root=root, path="person.primitives.b.c", list_to_items=False).list() == \
           [root["person"]["primitives"][3]["b"]["c"], root["person"]["primitives"][4]["b"]["c"]]

    # These are two separate branches, to list_to_items should have no effect
    assert PathVisitor(root=root, path="person.primitives.b.c.ping").list() == \
           ["pong1", "pong2"]
    assert PathVisitor(root=root, path="person.primitives.b.c.ping", list_to_items=False).list() == \
           ["pong1", "pong2"]

    assert PathVisitor(root=root, path="person.primitives.b.c.ping.", list_to_items=False).list() == []
    assert PathVisitor(root=root, path="person.primitives.b.c.ping.lalala", list_to_items=False).list() == []


def test_nicks():
    assert PathVisitor(root=root, path="person.nicks.short").list() == \
           [root["person"]["nicks"][0]["short"],
            root["person"]["nicks"][1]["short"],
            root["person"]["nicks"][2]["short"]]

    assert PathVisitor(root=root, path="person.nicks.short.roz").list() == []

    assert PathVisitor(root=root, path="person.nicks.long").list() == \
           [root["person"]["nicks"][3]["long"]]

    # noinspection PyUnresolvedReferences
    assert PathVisitor(root=root, path="person.nicks.long.heya").list() == \
           [*root["person"]["nicks"][3]["long"]["heya"]]
    # noinspection PyUnresolvedReferences
    assert PathVisitor(root=root, path="person.nicks.long.heya", list_to_items=False).list() == \
           [root["person"]["nicks"][3]["long"]["heya"]]


def test_simple_visit():
    def collect(v):
        results.append(v)

    results = []
    d = root_clone()
    PathVisitor(root=d, path="person.name.last").visit(collect)
    PathVisitor(root=d, path="person.name.first").visit(collect)
    assert results == ["roz", "elad"]


def test_list_visit():
    def collect(v):
        results.append(v)

    # Also uses visit() twice on same object
    results = []
    p = PathVisitor(root=root, path="person.primitives.b.c")
    p.visit(collect)
    p.visit(collect)
    # noinspection PyUnresolvedReferences
    assert results == [*root["person"]["primitives"][3]["b"]["c"], *root["person"]["primitives"][4]["b"]["c"],
                       *root["person"]["primitives"][3]["b"]["c"], *root["person"]["primitives"][4]["b"]["c"]]


def test_dont_modify():
    d = root_clone()
    PathVisitor(root=d, path="person.nicks").visit(lambda v: "123")
    PathVisitor(root=d, path="person.name.first").visit(lambda v: "mosh")
    PathVisitor(root=d, path="person.name.last").visit(lambda v: False)
    assert dict_equals_root(d)


def test_simple_modify():
    d = root_clone()
    p = PathVisitor(root=d, path="person.name.last", modifiable=True)
    p.visit(lambda v: "newname")
    assert d['person']['name']['last'] == "newname"
    p.visit(lambda v: "roz")
    assert dict_equals_root(d)


def test_modify_complex():
    d = root_clone()
    wp = PathVisitor(root=d, path="person.name.last", modifiable=True)
    # Replace primitive with dict
    wp.visit(lambda v: {"a": "aaa", "b": "bbb"})
    assert PathVisitor(root=d, path="person.name.last.b").list() == ["bbb"]

    wp.visit(lambda v: "roz")
    assert dict_equals_root(d)  # After 1

    # Replace dict with dict
    wp.visit(lambda v: {"a2": "aaa2", "b2": "bbb2"})
    assert PathVisitor(root=d, path="person.name.last.b2").list() == ["bbb2"]
    assert PathVisitor(root=d, path="person.name.last.b").list() == []

    wp.visit(lambda v: "roz")
    assert dict_equals_root(d)  # After 2

    # Replace dict with list
    wp.visit(lambda v: ["a", "b", "c", {"d": "e"}])
    assert wp.list() == ["a", "b", "c", {"d": "e"}]
    assert PathVisitor(root=d, path="person.name.last", list_to_items=False).list() == [["a", "b", "c", {"d": "e"}]]
    assert PathVisitor(root=d, path="person.name.last.d").list() == ["e"]

    # Replace some list elements
    wp.visit(lambda v: "roz" if v in ["b", "c"] else None)
    assert PathVisitor(root=d, path="person.name.last").list() == ["a", "roz", "roz", {"d": "e"}]

    # Replace inner list to dict
    wp.visit(lambda v: ["d", "e"] if v == {"d": "e"} else None)
    assert PathVisitor(root=d, path="person.name.last").list() == ["a", "roz", "roz", ["d", "e"]]

    # ...and back
    wp.visit(lambda v: {"d", "e"} if v == ["d", "e"] else None)
    assert PathVisitor(root=d, path="person.name.last").list() == ["a", "roz", "roz", {"d", "e"}]

    # Now return from list to original primitive value. For that we don't want the visitor to iterate list items,
    # but return the list object
    PathVisitor(root=d, path="person.name.last", modifiable=True, list_to_items=False).\
        visit(lambda v: "roz" if isinstance(v, list) else "wtf")
    assert dict_equals_root(d)  # After 3



"""
Implementation of queries (conditions, aggregations, funnel) with Pandas, over a single DataFrame.

The QueryEngine is agnostic to the wider system in which it runs - it is handed a DataFrame object and an already
validated query (with any shorthand notations and ommitted values expanded to the elaborate version), and does its work.
It should not import any code from the invoker/worker layers, and should remain testable as a stand-alone component.

TODO backlog get rid of any conditions/aggregations that are using composable strings (df.query()),
 although the query schema is validated in various layers rather than any client input naively taken as is.
"""

#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import logging
import math
from dataclasses import dataclass
from typing import Dict, List, Callable, Any, Union, Optional, cast
import numpy as np
import pandas as pd
from pandas import DataFrame, Series, Index
from frocket.common.config import config
from frocket.common.helpers.pandas import filter_by_isin, add_column_by_value_map
from frocket.common.tasks.query import AggregationResult, QueryConditionsResult, FunnelResult, \
    QueryResult, AggregationType
from frocket.common.validation.consts import CONDITION_COLUMN_PREFIX, TARGET_TYPES_WITH_INCLUDE_ZERO, NUMERIC_OPERATORS
from frocket.common.validation.path_visitor import PathVisitor
from frocket.common.validation.relation_parser import RelationParser
from frocket.engine.relation_to_pandas import relation_to_pandas_query

logger = logging.getLogger(__name__)
INCLUDE_ZERO = 'includeZero'
SUPPORTED_CONDITION_TYPES = {'filter', 'filters', 'sequence'}
AGGREGATION_TOP_DEFAULT_COUNT = config.int('aggregations.top.default.count')
AGGREGATION_TOP_GRACE_FACTOR = config.float('aggregations.top.grace.factor')


@dataclass(frozen=True)
class ColumnAggregatorContext:
    column: str
    other_column: Optional[str]
    group_by_column: str
    timestamp_column: str


ColumnAggregatorFunc = Callable[[DataFrame, ColumnAggregatorContext],
                                Union[int, float, Series]]

AGGREGATION_FUNCTIONS: Dict[AggregationType, ColumnAggregatorFunc] = {
    AggregationType.COUNT: lambda df, ctx: df[ctx.column].notnull().sum().item(),
    AggregationType.COUNT_PER_VALUE: lambda df, ctx: df[ctx.column].value_counts(),
    AggregationType.GROUPS_PER_VALUE: lambda df, ctx: df.groupby(ctx.column, sort=False)[ctx.group_by_column].nunique(),
    AggregationType.SUM_PER_VALUE: lambda df, ctx: df.groupby(ctx.column, sort=False)[ctx.other_column].sum(),
    AggregationType.MEAN_PER_VALUE: lambda df, ctx: df.groupby(ctx.column, sort=False)[ctx.other_column].mean()
}

FilterFunc = Callable[[DataFrame], Series]


class QueryEngine:
    def __init__(self, group_by_column: str, timestamp_column: str):
        self._group_by_column = group_by_column
        self._timestamp_column = timestamp_column

    def run(self, df: DataFrame, full_query: Dict) -> QueryResult:
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Got query: {json.dumps(full_query, indent=2)}")

        # First stage: run query conditions, if any
        conditions = PathVisitor(full_query, 'query.conditions').list()
        if conditions:
            relation_query = self._build_relation_query(full_query)
            matching_groups = self._run_query_conditions(df, conditions, relation_query)
            rows_of_matching_groups = filter_by_isin(df=df, column=self._group_by_column, values=matching_groups)
        else:
            # No query entry at all / empty or no conditions array means: all group IDs match, and thus all rows
            matching_groups = df[self._group_by_column].unique()
            rows_of_matching_groups = df

        # Second stage: run any aggregations defined over the query (regardless of whether there's also a funnel)
        query_aggregations = PathVisitor(full_query, 'query.aggregations').list()
        aggregation_results = \
            self._run_aggregations(rows_of_matching_groups, query_aggregations) if query_aggregations else None

        # Wrapping up the query part
        query_result = QueryConditionsResult(matching_groups=len(matching_groups),
                                             matching_group_rows=len(rows_of_matching_groups),
                                             aggregations=aggregation_results)

        # Third stage: run funnel, if defined - over rows of all groups matching the query conditions
        funnel = full_query.get('funnel', None)
        funnel_result = self._run_funnel(rows_of_matching_groups, funnel) if funnel else None

        return QueryResult(query_result, funnel_result)

    @staticmethod
    def _build_relation_query(full_query: dict) -> str:
        parsed_relation_elements = RelationParser(full_query).parse()
        return relation_to_pandas_query(parsed_relation_elements, CONDITION_COLUMN_PREFIX)

    def _run_query_conditions(self, df: DataFrame, conditions: List[dict], relation_query: str) -> np.ndarray:
        condition_results: List[Series] = []
        for i, condition in enumerate(conditions):
            filter_func = self._get_filter_function(condition)
            result_series = filter_func(df)
            condition_results.append(result_series.rename(CONDITION_COLUMN_PREFIX + str(i)))

        all_conditions_df = pd.concat(condition_results, axis=1, copy=False)
        all_conditions_df.fillna(False, inplace=True)

        groups_matching_relation = all_conditions_df.query(relation_query)
        return groups_matching_relation.index.values

    def _get_filter_function(self, condition: dict) -> FilterFunc:
        ftype_to_handler: Dict[str, FilterFunc] = {
            'filter': lambda df: self._run_condition(df, condition),
            'filters': lambda df: self._run_condition(df, condition),
            'sequence': lambda df: self._run_sequence_condition(df, condition)
        }

        # Ensure there's exxactly one of the supported filter types in the given condition
        found_filter_type = [k for k in condition.keys() if k in SUPPORTED_CONDITION_TYPES]
        assert len(found_filter_type) == 1
        return ftype_to_handler[found_filter_type[0]]

    # noinspection PyUnusedLocal
    def _run_condition(self, df: DataFrame, condition: Dict) -> Series:
        # First stage: run the filter query and group - only groups with one or more matching rows will remain
        filter_query_string = QueryEngine._build_query_string(condition)
        logger.debug(f"Running filter query: '{filter_query_string}' and grouping by '{self._group_by_column}'")
        matching_groups_df = df.query(filter_query_string).groupby(self._group_by_column, sort=False)

        # Second stage: run target
        target = condition['target']
        assert target['op'] in NUMERIC_OPERATORS
        assert type(target['value']) in [int, float]
        target_type = target['type']
        target_includes_zero = condition.get(INCLUDE_ZERO, False)
        if target_includes_zero:
            assert target_type in TARGET_TYPES_WITH_INCLUDE_ZERO

        if target_type == 'sum':
            group_sums = matching_groups_df[target['column']].sum()
            target_expression = f"group_sums {target['op']} {target['value']}"
            logger.debug(f"Evaluating target: '{target_expression}'. Group sums are for column '{target['column']}'")
            result_series = eval(target_expression)  # TODO backlog move to numexpr engine?
            result_series = result_series[result_series]  # Filter the series to rows whose value is True
            logger.debug(f"Group with sum matching target: {len(result_series)}")

        elif target_type == 'count':
            group_sizes = matching_groups_df.size()
            target_expression = f"group_sizes {target['op']} {target['value']}"
            logger.debug(f"Evaluating target: {target_expression}")
            result_series = eval(target_expression)
            result_series = result_series[result_series]  # Filter the series to rows whose value is True
            logger.debug(f"Group with size matching target: {len(result_series)} (no zero-sized groups)")

            if target_includes_zero:
                matching_ids = list(matching_groups_df.indices.keys())
                all_group_ids = df[self._group_by_column].unique()
                ids_wo_match_array = all_group_ids[~np.isin(all_group_ids, matching_ids)]
                ids_without_match = Series(data=True, index=ids_wo_match_array)
                logger.debug(f"Group IDs with zero metching rows: {len(ids_without_match)}")

                result_series = result_series.append(ids_without_match, verify_integrity=False)
                logger.debug(f"Group IDs either matching target or zero matches: {len(result_series)}")
        else:
            raise Exception(f"Uknown target type: {target_type}")

        return result_series

    @staticmethod
    def _build_query_string(condition: dict) -> str:
        if 'filter' in condition:
            return QueryEngine._single_filter_query_string(condition['filter'])
        elif 'filters' in condition:
            return QueryEngine._multi_filter_query_string(condition['filters'])
        else:
            raise Exception(f"Don't know how to handle condition: {condition}")

    @staticmethod
    def _single_filter_query_string(f: dict) -> str:
        # Back-ticks around column name allows names which aren't valid Python identifiers
        assert '`' not in f['column']  # Sanity/safety
        safe_column_name = f"`{f['column']}`"

        # Add quotes around string values
        value = f['value']
        if type(value) is str:
            value = value.replace("'", "\\'")
            value = "'" + value + "'"

        if f['op'] in ('contains', 'not contains'):
            return f"{'~' if f['op'].startswith('not') else ''}{safe_column_name}.str.contains({value}, \
                    case=True, na=False, regex=False)"
        else:
            assert f['op'] in NUMERIC_OPERATORS
            return f"{safe_column_name} {f['op']} {value}"

    @staticmethod
    def _multi_filter_query_string(filters: List[dict]) -> str:
        single_queries = [QueryEngine._single_filter_query_string(f) for f in filters]
        return ' & '.join(single_queries)  # Relation between elements in 'filters' is currently only AND

    def _run_sequence_condition(self, df: DataFrame, condition: dict) -> Series:
        last_step_group_ids = self._run_sequence(df, condition['sequence'], return_all_steps=False)[0]
        return pd.Series(data=True, index=last_step_group_ids)

    def _run_sequence(self, df: DataFrame, sequence: List[dict], return_all_steps: bool = True) -> List[Index]:
        def column_for_min_timestamp(step_idx: int) -> str:
            return f"ts_of_event_{step_idx}"

        step_queries = [QueryEngine._build_query_string(f) for f in sequence]
        df_with_step_columns = df.copy(deep=False)
        users_who_passed_each_step = []

        for i, query_string in enumerate(step_queries):
            if i == 0:
                query_to_run = query_string
            else:
                query_to_run = f"{self._timestamp_column} > {column_for_min_timestamp(i - 1)} & {query_string}"

            rows_matching_step = df_with_step_columns.query(query_to_run)
            min_ts_per_matching_group = \
                rows_matching_step.groupby(self._group_by_column, sort=False)[self._timestamp_column].min()

            is_last_step = (i == len(step_queries) - 1)
            if return_all_steps or is_last_step:
                users_who_passed_each_step.append(min_ts_per_matching_group.index)

            if not is_last_step:
                add_column_by_value_map(df=df_with_step_columns,
                                        keys_column=self._group_by_column,
                                        values_map_series=min_ts_per_matching_group,
                                        new_column=column_for_min_timestamp(i))

        return users_who_passed_each_step

    def _run_aggregations(self,
                          df: DataFrame,
                          aggregations: Optional[List[dict]]) -> Optional[List[AggregationResult]]:
        if aggregations is None:
            return None
        elif not aggregations:
            return []

        result = []
        for agg in aggregations:
            agg_type = cast(AggregationType, AggregationType.from_camelcase(agg['type']))
            if agg_type.value_is_dict:
                top_count = agg.get('top', AGGREGATION_TOP_DEFAULT_COUNT)
                actual_top_count = math.ceil(top_count * AGGREGATION_TOP_GRACE_FACTOR)
            else:
                top_count = None
                actual_top_count = None

            if len(df) == 0:
                value = {} if agg_type.value_is_dict else 0
            else:
                ctx = ColumnAggregatorContext(column=agg['column'],
                                              other_column=agg.get('otherColumn', None),
                                              group_by_column=self._group_by_column,
                                              timestamp_column=self._timestamp_column)
                aggr_func = AGGREGATION_FUNCTIONS[agg_type]
                aggr_result = aggr_func(df, ctx)
                if agg_type.value_is_dict:
                    assert type(aggr_result) is Series
                    top_dict = aggr_result.nlargest(actual_top_count).to_dict()
                    value = {str(k): v for k, v in top_dict.items()}
                else:
                    assert type(aggr_result) in [int, float]
                    value = aggr_result

            result.append(AggregationResult(column=agg['column'],
                                            type=agg['type'],
                                            name=agg.get('name', None),
                                            value=value,
                                            top=top_count))

        return result

    def _run_funnel(self, df: DataFrame, funnel: Dict[str, Any]) -> FunnelResult:
        # First stage: calculate the group IDs and all-their-rows for each step in the funnel
        groups_per_step = self._run_sequence(df, funnel['sequence'])
        all_rows_per_step = [filter_by_isin(df, self._group_by_column, values=groups)
                             for groups in groups_per_step]

        # Second stage: if there are aggregations to run after each step, run 'em now for all steps
        step_aggregations = PathVisitor(funnel, 'stepAggregations').list()
        if step_aggregations:
            step_agg_results = [self._run_aggregations(step_df, step_aggregations) for
                                step_df in all_rows_per_step]
        else:
            step_agg_results = None

        # Third stage: if the are aggregations to run after the full funnel is completed, run
        end_aggregations = PathVisitor(funnel, 'endAggregations').list()
        if end_aggregations:
            end_agg_results = self._run_aggregations(all_rows_per_step[-1], end_aggregations)
        else:
            end_agg_results = None

        per_step_results = [
            QueryConditionsResult(matching_groups=len(groups_per_step[i]),
                                  matching_group_rows=len(all_rows_per_step[i]),
                                  aggregations=step_agg_results[i] if step_agg_results else None)
            for i in range(len(all_rows_per_step))
        ]
        return FunnelResult(sequence=per_step_results, end_aggregations=end_agg_results)



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict, Type
from frocket.common.tasks.base import BaseTaskRequest
from frocket.common.tasks.registration import RegistrationTaskRequest
from frocket.common.tasks.query import QueryTaskRequest
from frocket.worker.runners.base_task_runner import BaseTaskRunner
from frocket.worker.runners.query_task_runner import QueryTaskRunner
from frocket.worker.runners.registration_task_runner import RegistrationTaskRunner

REGISTERED_RUNNERS: Dict[Type[BaseTaskRequest], Type[BaseTaskRunner]] = {
    QueryTaskRequest: QueryTaskRunner,
    RegistrationTaskRequest: RegistrationTaskRunner
}



"""
This is the "Funnel Rocket" frontend API - wrappedby the CLI & API server, and may be embeddable in other apps.
Clients are not expected to bypass this API (call the datastore directly, initialize an invoker, etc.)
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import concurrent.futures
import logging
import time
from typing import List, Optional, cast, Union
from frocket.common.config import config
from frocket.common.dataset import DatasetInfo, DatasetShortSchema, DatasetSchema, DatasetPartsInfo
from frocket.common.tasks.registration import RegistrationJobResult, RegisterArgs, UnregisterApiResult
from frocket.common.tasks.query import QueryJobResult
from frocket.common.tasks.async_tracker import AsyncJobTracker, AsyncJobStatusUpdater
from frocket.common.validation.query_validator import QueryValidator
from frocket.common.validation.result import QueryValidationResult
from frocket.datastore.registered_datastores import get_datastore
from frocket.invoker.jobs.query_job import QueryJob
from frocket.invoker.jobs.registration_job import RegistrationJob
from frocket.invoker.impl.registered_invokers import new_invoker

logger = logging.getLogger(__name__)
executor = concurrent.futures.ThreadPoolExecutor()

# TODO backlog allow configurable timeout per job type (async or not)
ASYNC_MAX_WAIT = config.int("invoker.run.timeout") * 1.1  # Adding a bit of grace around the invoker


def _unregister_safety_interval() -> int:
    """How long after a dataset was last used to block unregister (can be set to zero, or overidden with force=True)."""
    interval = config.get('unregister.last.used.interval', None)
    if not interval:  # Not defined, or empty string (explicit '0' is truthy)
        interval = config.int('invoker.run.timeout') * 2
    else:
        interval = int(interval)
    return interval


def register_dataset(args: RegisterArgs) -> RegistrationJobResult:
    request_builder = RegistrationJob(args)
    invoker = new_invoker(request_builder)
    result = cast(RegistrationJobResult, invoker.run())
    logger.info(f"Registration {'successful' if result.success else f'failed! {result.error_message}'}")
    return result


def register_dataset_async(args: RegisterArgs, set_max_wait: bool = True) -> AsyncJobTracker:
    """The async version starts the invoker in a separate thread and then returns, handing back
    an AsyncJobTracker to poll for progress/completion."""
    def worker(register_args, async_status):
        invoker = new_invoker(RegistrationJob(register_args))
        return invoker.run(async_status)

    async_status = AsyncJobStatusUpdater(max_wait=(ASYNC_MAX_WAIT if set_max_wait else None))
    executor.submit(worker, args, async_status)
    logger.info(f"Submitted async registration for dataset named {args.name} in basepath {args.basepath}")
    return async_status


def get_dataset(name: str, throw_if_missing: bool = False) -> Optional[DatasetInfo]:
    dataset = get_datastore().dataset_info(name)
    if not dataset and throw_if_missing:
        raise Exception(f"Dataset '{name}' not found")
    return dataset


def get_dataset_schema(dataset: DatasetInfo, full: bool = False) -> Union[DatasetSchema, DatasetShortSchema]:
    return get_datastore().schema(dataset) if full else get_datastore().short_schema(dataset)


def get_dataset_parts(dataset: DatasetInfo) -> DatasetPartsInfo:
    return get_datastore().dataset_parts_info(dataset)


def unregister_dataset(name: str, force: bool = False) -> UnregisterApiResult:
    dataset = get_dataset(name=name)
    if not dataset:
        return UnregisterApiResult(success=True, error_message=None,
                                   dataset_found=False, dataset_last_used=None)

    datastore = get_datastore()
    last_used = datastore.last_used(dataset)
    if last_used:
        time_since_used = int(time.time() - last_used)
        safety_interval = _unregister_safety_interval()
        message = f"Dataset was last used {time_since_used} seconds ago, which is less than safety interval " \
                  f"{safety_interval}. Use the 'force' parameter to unregister anyway."
        if safety_interval > time_since_used and not force:
            return UnregisterApiResult(success=False, error_message=message,
                                       dataset_found=True, dataset_last_used=last_used)

    get_datastore().remove_dataset_info(name)
    return UnregisterApiResult(success=True, error_message=None,
                               dataset_found=True, dataset_last_used=last_used)


def expand_and_validate_query(dataset: DatasetInfo, query: dict) -> QueryValidationResult:
    short_schema = get_dataset_schema(dataset)
    return QueryValidator(query, dataset, short_schema).expand_and_validate()


def _build_query_job(dataset: DatasetInfo,
                     query: dict,
                     validation_result: QueryValidationResult) -> QueryJob:
    """If the query was already validated, skip re-validating."""
    if validation_result:
        assert validation_result.success
        assert query in [validation_result.source_query, validation_result.expanded_query]
    else:
        validation_result = expand_and_validate_query(dataset, query)
        if not validation_result.success:
            raise Exception(f"Query validation failed: {validation_result.error_message}")

    get_datastore().mark_used(dataset)
    dataset_parts = get_datastore().dataset_parts_info(dataset)
    short_schema = get_datastore().short_schema(dataset)
    return QueryJob(dataset, dataset_parts, short_schema,
                    validation_result.expanded_query, validation_result.used_columns)


def run_query(dataset: DatasetInfo,
              query: dict,
              validation_result: QueryValidationResult = None) -> QueryJobResult:
    job_builder = _build_query_job(dataset, query, validation_result)
    invoker = new_invoker(job_builder)
    result = cast(QueryJobResult, invoker.run())
    if result.success:
        logger.info("Query completed successfully")
    else:
        logger.error(f"Query failed with message: {result.error_message}")
    return result


def run_query_async(dataset: DatasetInfo,
                    query: dict,
                    set_max_wait: bool = True,
                    validation_result: QueryValidationResult = None) -> AsyncJobTracker:
    """The async version starts the invoker in a separate thread and then returns, handing back
    an AsyncJobTracker to poll for progress/completion."""
    def worker(job_builder, async_status):
        invoker = new_invoker(job_builder)
        return invoker.run(async_status)

    job_builder = _build_query_job(dataset, query, validation_result)
    async_status = AsyncJobStatusUpdater(max_wait=(ASYNC_MAX_WAIT if set_max_wait else None))
    executor.submit(worker, job_builder, async_status)
    logger.info(f"Submitted async query for dataset '{dataset.id.name}'")
    return async_status


def list_datasets() -> List[DatasetInfo]:
    datasets = get_datastore().datasets()
    return datasets



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from dataclasses import dataclass
from typing import Optional, List, cast, Dict
from frocket.common.serializable import SerializableDataClass
from frocket.common.validation.error import ValidationErrorKind, QueryValidationError
from frocket.common.validation.relation_parser import RBaseElement


@dataclass(frozen=True)
class QueryValidationResult(SerializableDataClass):
    success: bool
    source_query: dict
    error_message: Optional[str] = None
    error_kind: Optional[ValidationErrorKind] = None
    expanded_query: Optional[dict] = None
    # TODO backlog support non-critical warning/hints to user (e.g. conditions unused by relation expression)
    warnings: Optional[List[str]] = None
    used_columns: Optional[List[str]] = None
    used_conditions: Optional[List[str]] = None
    named_conditions: Optional[Dict[str, int]] = None
    relation_elements: Optional[List[RBaseElement]] = None

    @staticmethod
    def from_exception(e: Exception, source_query: dict):
        if type(e) is QueryValidationError:
            error_kind = cast(QueryValidationError, e).kind
        else:
            error_kind = ValidationErrorKind.UNEXPECTED
        return QueryValidationResult(success=False, error_message=str(e), error_kind=error_kind,
                                     source_query=source_query)
